import pathlib
import cv2
import numpy as np
import typing
import enum

from glassesTools import annotation, data_files, drawing, intervals, marker, ocv, timestamps, transforms, _has_GUI
if _has_GUI:
    from .gui import video_player
else:
    # stub out video_player as a class so type annotations below do not fail
    class video_player:
        @property
        def GUI(self) -> typing.Any: ...

class Pose:
    # description of tsv file used for storage
    _columns_compressed = {'frame_idx':1,
                           'pose_N_markers': 1, 'pose_reprojection_error': 1, 'pose_R_vec': 3, 'pose_T_vec': 3,
                           'homography_N_markers': 1, 'homography_mat': 9}
    _non_float          = {'frame_idx': int, 'pose_ok': bool, 'pose_N_markers': int, 'homography_N_markers': int}

    def __init__(self,
                 frame_idx              : int,
                 pose_N_markers         : int       = 0,
                 pose_reprojection_error: float     = -1.,
                 pose_R_vec             : np.ndarray= None,
                 pose_T_vec             : np.ndarray= None,
                 homography_N_markers   : int       = 0,
                 homography_mat         : np.ndarray= None):
        self.frame_idx              : int         = frame_idx
        # pose
        self.pose_N_markers         : int         = pose_N_markers        # number of ArUco markers this pose estimate is based on. 0 if failed
        self.pose_reprojection_error: float       = pose_reprojection_error
        self.pose_R_vec             : np.ndarray  = pose_R_vec
        self.pose_T_vec             : np.ndarray  = pose_T_vec
        # homography
        self.homography_N_markers   : int         = homography_N_markers  # number of ArUco markers this homongraphy estimate is based on. 0 if failed
        self.homography_mat         : np.ndarray  = homography_mat.reshape(3,3) if homography_mat is not None else homography_mat

        # internals
        self._RMat              = None
        self._RtMat             = None
        self._plane_normal      = None
        self._plane_point       = None
        self._RMatInv           = None
        self._RtMatInv          = None
        self._i_homography_mat  = None

    def pose_successful(self):
        return self.pose_N_markers>0
    def homography_successful(self):
        return self.homography_N_markers>0

    def draw_frame_axis(self, img, camera_params: ocv.CameraParams, arm_length, thickness, sub_pixel_fac, position = [0.,0.,0.]):
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or not camera_params.has_intrinsics():
            return
        drawing.openCVFrameAxis(img, camera_params, self.pose_R_vec, self.pose_T_vec, arm_length, thickness, sub_pixel_fac, position)

    def cam_frame_to_world(self, point: np.ndarray):
        # NB: world frame is in the plane's coordinate system
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(point)):
            return np.full((3,), np.nan)

        if self._RtMatInv is None:
            if self._RMatInv is None:
                if self._RMat is None:
                    self._RMat = cv2.Rodrigues(self.pose_R_vec)[0]
                self._RMatInv = self._RMat.T
            self._RtMatInv = np.hstack((self._RMatInv,np.matmul(-self._RMatInv,self.pose_T_vec.reshape(3,1))))

        return np.matmul(self._RtMatInv,np.append(np.array(point),1.).reshape((4,1))).flatten()

    def world_frame_to_cam(self, point: np.ndarray):
        # NB: world frame is in the plane's coordinate system
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(point)):
            return np.full((3,), np.nan)

        if self._RtMat is None:
            if self._RMat is None:
                self._RMat = cv2.Rodrigues(self.pose_R_vec)[0]
            self._RtMat = np.hstack((self._RMat, self.pose_T_vec.reshape(3,1)))

        return np.matmul(self._RtMat,np.append(np.array(point),1.).reshape((4,1))).flatten()

    def plane_to_cam_pose(self, point_plane: np.ndarray, camera_params: ocv.CameraParams) -> np.ndarray:
        # from location on plane (2D) to location on camera image (2D)
        if (self.pose_R_vec is None) or (self.pose_T_vec is None):
            return np.full((2,), np.nan)
        return transforms.project_points(point_plane, camera_params, rot_vec=self.pose_R_vec, trans_vec=self.pose_T_vec).flatten()

    def cam_to_plane_pose(self, point: np.ndarray, camera_params: ocv.CameraParams) -> tuple[np.ndarray,np.ndarray]:
        # from location on camera image (2D) to location on plane (2D)
        # NB: also returns intermediate result (intersection with plane in camera space)
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(point)) or not camera_params.has_intrinsics():
            return np.full((2,), np.nan), np.full((3,), np.nan)

        g3D = transforms.unproject_points(point, camera_params)

        # find intersection of 3D gaze with plane
        pos_cam = self.vector_intersect(g3D)  # default vec origin (0,0,0) because we use g3D from camera's view point

        # above intersection is in camera space, turn into world space to get position on plane
        (x,y,z) = self.cam_frame_to_world(pos_cam) # z should be very close to zero
        return np.asarray([x, y]), pos_cam

    def plane_to_cam_homography(self, point: np.ndarray, camera_params: ocv.CameraParams) -> np.ndarray:
        # from location on plane (2D) to location on camera image (2D)
        if self.homography_mat is None:
            return np.full((2,), np.nan)

        if self._i_homography_mat is None:
            self._i_homography_mat = np.linalg.inv(self.homography_mat)
        out = transforms.apply_homography(point, self._i_homography_mat).flatten()
        if camera_params.has_intrinsics():
            out = transforms.distort_points(out, camera_params).flatten()
        return out

    def cam_to_plane_homography(self, point_cam: np.ndarray, camera_params: ocv.CameraParams) -> np.ndarray:
        # from location on camera image (2D) to location on plane (2D)
        if self.homography_mat is None:
            return np.full((2,), np.nan)

        if camera_params.has_intrinsics():
            point_cam = transforms.undistort_points(point_cam, camera_params).flatten()
        return transforms.apply_homography(point_cam, self.homography_mat).flatten()

    def get_origin_on_image(self, camera_params: ocv.CameraParams) -> np.ndarray:
        if self.pose_successful() and camera_params.has_intrinsics():
            a = self.plane_to_cam_pose(np.zeros((1,3)), camera_params)
        elif self.homography_successful():
            a = self.plane_to_cam_homography(np.zeros((1,2)), camera_params)
        else:
            a = np.full((2,), np.nan)
        return a

    def vector_intersect(self, vector: np.ndarray, origin = np.array([0.,0.,0.])):
        if (self.pose_R_vec is None) or (self.pose_T_vec is None) or np.any(np.isnan(vector)):
            return np.full((3,), np.nan)

        if self._plane_normal is None:
            if self._RtMat is None:
                if self._RMat is None:
                    self._RMat = cv2.Rodrigues(self.pose_R_vec)[0]
                self._RtMat = np.hstack((self._RMat, self.pose_T_vec.reshape(3,1)))

            # get poster normal
            self._plane_normal = self._RMat[:,2]     # equivalent to: np.matmul(self._RMat, np.array([0., 0., 1.]))
            # get point on poster (just use origin)
            self._plane_point  = self._RtMat[:,3]    # equivalent to: np.matmul(self._RtMat, np.array([0., 0., 0., 1.]))

        # normalize vector
        vector /= np.linalg.norm(vector)

        # find intersection of 3D gaze with poster
        return transforms.intersect_plane_ray(self._plane_normal, self._plane_point, vector.flatten(), origin.flatten())


def read_dict_from_file(fileName:str|pathlib.Path, episodes:list[list[int]]=None) -> dict[int,Pose]:
    return data_files.read_file(fileName,
                                Pose, True, True, False, False,
                                episodes=episodes)[0]

def write_list_to_file(poses: list[Pose], fileName:str|pathlib.Path, skip_failed=False):
    data_files.write_array_to_file(poses, fileName,
                                    Pose._columns_compressed,
                                    skip_all_nan=skip_failed)

class Status(enum.Enum):
    Ok = enum.auto()
    Skip = enum.auto()
    Finished = enum.auto()

_T  = typing.TypeVar("_T")

class Estimator:
    def __init__(self, video_file: str|pathlib.Path, frame_timestamp_file: str|pathlib.Path|timestamps.VideoTimestamps, camera_calibration_file: str|pathlib.Path|ocv.CameraParams):
        self.video_ts   = frame_timestamp_file if isinstance(frame_timestamp_file,timestamps.VideoTimestamps) else timestamps.VideoTimestamps(frame_timestamp_file)
        self.video      = ocv.CV2VideoReader(video_file, self.video_ts.timestamps)
        self.cam_params = camera_calibration_file if isinstance(camera_calibration_file,ocv.CameraParams) else ocv.CameraParams.read_from_file(camera_calibration_file)

        self.plane_functions    : dict[str, typing.Callable[[str,int,np.ndarray,ocv.CameraParams], tuple[np.ndarray,np.ndarray]|None]] = {}
        self.plane_intervals    : dict[str, list[int]|list[list[int]]]                                  = {}
        self.plane_visualizers  : dict[str, typing.Callable[[str,int,np.ndarray,np.ndarray], None]|None]= {}

        self.individual_marker_functions    : dict[_T, typing.Callable[[_T,int,np.ndarray,ocv.CameraParams], tuple[np.ndarray,np.ndarray|None|None]]] = {}
        self.individual_marker_intervals    : dict[_T, list[int]|list[list[int]]]                                   = {}
        self.individual_marker_visualizers  : dict[_T, typing.Callable[[_T,int,np.ndarray,np.ndarray], None]|None]  = {}

        self.extra_proc_functions   : dict[str, typing.Callable[[str,int,np.ndarray,ocv.CameraParams,typing.Any], typing.Any]]   = {}
        self.extra_proc_intervals   : dict[str, list[int]|list[list[int]]]                                  = {}
        self.extra_proc_parameters  : dict[str, dict[str,typing.Any]]                                       = {}
        self.extra_proc_visualizers : dict[str, typing.Callable[[str,int,np.ndarray,typing.Any], None]|None]= {}

        self._cache: tuple[Status, dict[str, Pose], dict[_T, marker.Pose], dict[str, list[int, typing.Any]], tuple[np.ndarray, int, float]] = None  # self._cache[4][1] is frame number

        self.gui                    : video_player.GUI          = None
        self.has_gui                                            = False
        self.allow_early_exit                                   = True
        self.progress_updater       : typing.Callable[[], None] = None

        self.do_visualize                                       = False
        self.sub_pixel_fac                                      = 8
        self.plane_axis_arm_length                              = 25
        self.individual_marker_axis_arm_length                  = 25
        self.show_extra_processing_output                       = True

        self._first_frame                                       = True

    def __del__(self):
        if self.has_gui:
            self.gui.stop()

    def add_plane(self, plane: str,
                  plane_function: typing.Callable[[str,int,np.ndarray,ocv.CameraParams], tuple[np.ndarray,np.ndarray]|None],
                  processing_intervals: list[int]|list[list[int]]=None,
                  plane_visualizer: typing.Callable[[str,int,np.ndarray,np.ndarray], None]=None):
        if not self._first_frame:
            raise RuntimeError(f'You cannot register planes once video processing has started')
        if plane in self.plane_functions:
            raise ValueError(f'Cannot register the plane "{plane}", it is already registered')
        self.plane_functions[plane]     = plane_function
        self.plane_intervals[plane]     = processing_intervals
        self.plane_visualizers[plane]   = plane_visualizer

    def add_individual_marker(self, key: _T,
                              individual_marker_function: typing.Callable[[_T,int,np.ndarray,ocv.CameraParams], tuple[np.ndarray,np.ndarray|None|None]],
                              processing_intervals: list[int]|list[list[int]]=None,
                              individual_marker_visualizer: typing.Callable[[str,int,np.ndarray,np.ndarray], None]=None):
        if not self._first_frame:
            raise RuntimeError(f'You cannot register individual markers once video processing has started')
        if key in self.individual_marker_functions:
            raise ValueError(f'Cannot register the individual marker {key}, it is already registered')
        self.individual_marker_functions[key]   = individual_marker_function
        self.individual_marker_intervals[key]   = processing_intervals
        self.individual_marker_visualizers[key] = individual_marker_visualizer

    def register_extra_processing_fun(self,
                                      name: str,
                                      func: typing.Callable[[str,int,np.ndarray,ocv.CameraParams,typing.Any], typing.Any],
                                      processing_intervals: list[int]|list[list[int]],
                                      func_parameters: dict[str],
                                      visualizer: typing.Callable[[str,int,np.ndarray,typing.Any], None]):
        if not self._first_frame:
            raise RuntimeError(f'You cannot register extra processing functions once video processing has started')
        if name in self.extra_proc_functions:
            raise ValueError(f'Cannot register the extra processing function "{name}", it is already registered')
        self.extra_proc_functions[name] = func
        self.extra_proc_intervals[name] = processing_intervals
        self.extra_proc_parameters[name]= func_parameters
        self.extra_proc_visualizers[name]= visualizer

    def attach_gui(self, gui: video_player.GUI, episodes: dict[annotation.Event, list[int]] = None, window_id: int = None):
        self.gui            = gui
        self.has_gui        = self.gui is not None
        self.do_visualize   = self.has_gui

        if self.has_gui:
            self.gui.set_show_timeline(True, self.video_ts, episodes, window_id)

    def set_allow_early_exit(self, allow_early_exit: bool):
        # if False, processing will not stop because last frame with a defined plane or extra processing is reached
        self.allow_early_exit = allow_early_exit

    def set_progress_updater(self, progress_updater: typing.Callable[[], None]):
        self.progress_updater = progress_updater

    def set_visualize_on_frame(self, do_visualize: bool):
        self.do_visualize = do_visualize

    def get_video_info(self) -> tuple[int, int, float]:
        return int(self.video.get_prop(cv2.CAP_PROP_FRAME_WIDTH)), \
               int(self.video.get_prop(cv2.CAP_PROP_FRAME_HEIGHT)), \
                   self.video.get_prop(cv2.CAP_PROP_FPS)

    def estimate_pose(self, object_points, img_points) -> tuple[int, np.ndarray, np.ndarray, float]:
        # NB: N_markers also flags success of the pose estimation. Also 0 if not successful or not possible (missing intrinsics)
        N_markers, R_vec, T_vec, reprojection_error = 0, None, None, -1.
        if object_points is None or not self.cam_params.has_intrinsics():
            return N_markers, R_vec, T_vec, reprojection_error

        if self.cam_params.has_opencv_camera():
            n_solutions, R_vec, T_vec, reprojection_error = \
                cv2.solvePnPGeneric(object_points, img_points, self.cam_params.camera_mtx, self.cam_params.distort_coeffs, np.empty(1), np.empty(1))
            if n_solutions:
                N_markers = int(object_points.shape[0]/4)
            reprojection_error = reprojection_error[0][0]
        else:
            # we have a camera not supported by OpenCV
            # undistort points and project to a identity camera space, so we can use opencv functionality
            points_w  = transforms.unproject_points(img_points, self.cam_params)
            points_cam= transforms.project_points(points_w, ocv.CameraParams(self.cam_params.resolution, np.identity(3), np.zeros((5,1))))
            n_solutions, R_vec, T_vec, _ = cv2.solvePnPGeneric(object_points, points_cam.reshape((-1,1,2)), np.identity(3), np.zeros((5,1)), np.empty(1), np.empty(1))
            # need to compute reprojection error ourselves, output of solvePnPGeneric is meaningless due to arbitrary camera point units
            if n_solutions:
                proj_points = transforms.project_points(object_points,self.cam_params, rot_vec=R_vec[0], trans_vec=T_vec[0])
                reprojection_error = cv2.norm(proj_points.astype('float32').reshape((-1,1,2)),img_points,cv2.NORM_L2) / np.sqrt(2*proj_points.shape[0])
            else:
                reprojection_error = np.nan
        if n_solutions:
            N_markers = int(object_points.shape[0]/4)
        return N_markers, R_vec[0], T_vec[0], reprojection_error

    def estimate_homography(self, object_points, img_points) -> tuple[np.ndarray, bool]:
        # NB: N_markers also flags success of the pose estimation. Also 0 if not successful or not possible (missing intrinsics)
        N_markers, H = 0, None
        if object_points is None:
            return N_markers, H

        # use undistorted marker corners if possible
        if self.cam_params.has_intrinsics():
            img_points = transforms.undistort_points(img_points.reshape((-1,2)),self.cam_params).reshape((-1,1,2))

        H = transforms.estimate_homography(object_points, img_points)
        if H is not None:
            N_markers = int(object_points.shape[0]/4)
        return N_markers, H

    # higher level functions for detecting + pose estimation
    def estimate_pose_and_homography(self, frame_idx, img_points, object_points) -> tuple[Pose, dict[str]]:
        pose = Pose(frame_idx)
        if img_points is not None and object_points is not None:
            # get camera pose
            pose.pose_N_markers, pose.pose_R_vec, pose.pose_T_vec, pose.pose_reprojection_error = \
                self.estimate_pose(object_points, img_points)

            # also get homography (direct image plane to plane in world transform)
            pose.homography_N_markers, pose.homography_mat = \
                self.estimate_homography(object_points, img_points)
        return pose

    def process_one_frame(self, wanted_frame_idx:int = None) -> tuple[Status, dict[str, Pose], dict[str, marker.Pose], dict[str, list[int, typing.Any]], tuple[np.ndarray, int, float]]:
        if self._first_frame and self.has_gui:
            self.gui.set_playing(True)

        if wanted_frame_idx is not None and self._cache is not None and self._cache[4][1]==wanted_frame_idx:
            return self._cache

        should_exit, frame, frame_idx, frame_ts = self.video.read_frame(report_gap=True, wanted_frame_idx=wanted_frame_idx)

        if should_exit or (self.allow_early_exit and \
            (
                (not self.plane_intervals             or intervals.beyond_last_interval(frame_idx, self.plane_intervals)) and \
                (not self.individual_marker_intervals or intervals.beyond_last_interval(frame_idx, self.individual_marker_intervals)) and \
                (not self.extra_proc_intervals        or intervals.beyond_last_interval(frame_idx, self.extra_proc_intervals))
            )):
            self._cache = Status.Finished, None, None, None, (None, None, None)
            return self._cache
        if self.progress_updater:
            self.progress_updater()

        if self.has_gui:
            if self._first_frame and frame is not None:
                self.gui.set_frame_size(frame.shape)
                self._first_frame = False

            requests = self.gui.get_requests()
            for r,_ in requests:
                if r=='exit':   # only requests we need to handle
                    self._cache = Status.Finished, None, None, None, (None, None, None)
                    return self._cache
                if r=='close':
                    self.has_gui = False
                    self.gui.stop()

        # check we're in a current interval, else skip processing
        # NB: have to spool through like this, setting specific frame to read
        # with cap.set(cv2.CAP_PROP_POS_FRAMES) doesn't seem to work reliably
        # for VFR video files
        planes_for_this_frame = [p for p in self.plane_functions if intervals.is_in_interval(frame_idx, self.plane_intervals[p])]
        indiv_markers_for_this_frame = [i for i in self.individual_marker_functions if intervals.is_in_interval(frame_idx, self.individual_marker_intervals[i])]
        extra_processing_for_this_frame = [e for e in self.extra_proc_functions if intervals.is_in_interval(frame_idx, self.extra_proc_intervals[e])]
        if frame is None or (not planes_for_this_frame and not indiv_markers_for_this_frame and not extra_processing_for_this_frame):
            # we don't have a valid frame or nothing to do, continue to next
            if self.has_gui:
                # do update timeline of the viewers
                self.gui.update_image(None, frame_ts/1000., frame_idx)
            self._cache = Status.Skip, None, None, None, (frame, frame_idx, frame_ts)
            return self._cache

        pose_out                : dict[str, Pose]                   = {}
        individual_marker_out   : dict[_T , marker.Pose]            = {}
        extra_processing_out    : dict[str, list[int, typing.Any]]  = {}
        if planes_for_this_frame:
            # detect fiducials
            plane_points: dict[str, tuple[np.ndarray,np.ndarray]] = {}
            for p in planes_for_this_frame:
                det_output = self.plane_functions[p](p, frame_idx, frame, self.cam_params)
                if det_output[0] is not None:
                    plane_points[p] = det_output
            # determine pose
            for p in plane_points:
                pose_out[p] = self.estimate_pose_and_homography(frame_idx, *plane_points[p])

        if indiv_markers_for_this_frame:
            # detect fiducials
            indiv_marker_points: dict[_T, tuple[np.ndarray,np.ndarray]] = {}
            for i in indiv_markers_for_this_frame:
                det_output = self.individual_marker_functions[i](i, frame_idx, frame, self.cam_params)
                if det_output[0] is not None:
                    indiv_marker_points[i] = det_output
            # determine pose, if wanted
            for i in indiv_marker_points:
                mpose = marker.Pose(frame_idx)
                if indiv_marker_points[i][1] is not None:   # object points may not be available (e.g. when marker size is not set)
                    # can only get marker pose if we have a calibrated camera (need intrinsics), else at least flag that marker was found
                    if self.cam_params.has_opencv_camera():
                        _, mpose.R_vec, mpose.T_vec = cv2.solvePnP(indiv_marker_points[i][1], indiv_marker_points[i][0], self.cam_params.camera_mtx, self.cam_params.distort_coeffs, flags=cv2.SOLVEPNP_IPPE_SQUARE)
                    elif self.cam_params.has_colmap_camera():
                        # we have a camera not supported by OpenCV
                        # undistort points and project to a identity camera space, so we can use opencv functionality
                        points_w  = transforms.unproject_points(indiv_marker_points[i][0],self.cam_params)
                        points_cam= transforms.project_points(points_w, ocv.CameraParams(self.cam_params.resolution, np.identity(3), np.zeros((5,1))))
                        cv2.solvePnP(indiv_marker_points[i][1], points_cam, np.identity(3), np.zeros((5,1)), flags=cv2.SOLVEPNP_IPPE_SQUARE)
                individual_marker_out[i] = mpose

        for e in extra_processing_for_this_frame:
            extra_processing_out[e] = [frame_idx, *self.extra_proc_functions[e](e, frame_idx, frame, self.cam_params, **self.extra_proc_parameters[e])]

        # now that all processing is done, handle visualization, if any
        if self.do_visualize:
            # first draw all detection output
            if planes_for_this_frame:
                for p in plane_points:
                    if self.plane_visualizers[p] is None:
                        continue
                    self.plane_visualizers[p](p, frame_idx, frame, plane_points[p][0])
            if indiv_markers_for_this_frame:
                for i in indiv_marker_points:
                    if self.individual_marker_visualizers[i] is None:
                        continue
                    self.individual_marker_visualizers[i](i, frame_idx, frame, indiv_marker_points[i][0])
            for e in extra_processing_for_this_frame:
                if self.show_extra_processing_output and self.extra_proc_visualizers[e]:
                    self.extra_proc_visualizers[e](e, frame, *extra_processing_out[e])

            # now also draw pose, if wanted
            if self.plane_axis_arm_length:
                for p in pose_out:
                    if pose_out[p].pose_successful():
                        pose_out[p].draw_frame_axis(frame, self.cam_params, self.plane_axis_arm_length, 3, sub_pixel_fac=self.sub_pixel_fac)
            if self.individual_marker_axis_arm_length:
                for i in individual_marker_out:
                    if individual_marker_out[i].pose_successful():
                        individual_marker_out[i].draw_frame_axis(frame, self.cam_params, self.individual_marker_axis_arm_length, self.sub_pixel_fac)

        if self.has_gui:
            self.gui.update_image(frame, frame_ts/1000., frame_idx)

        self._cache = Status.Ok, pose_out, individual_marker_out, extra_processing_out, (frame, frame_idx, frame_ts)
        return self._cache

    def process_video(self) -> tuple[dict[str, list[Pose]], dict[_T, list[marker.Pose]], dict[str, list[list[int, typing.Any]]]]:
        poses_out               : dict[str, list[Pose]]                 = {p:[] for p in self.plane_functions}
        individual_markers_out  : dict[_T, list[marker.Pose]]           = {i:[] for i in self.individual_marker_functions}
        extra_processing_out    : dict[str, list[list[int, typing.Any]]]= {e:[] for e in self.extra_proc_functions}
        while True:
            status, plane, individual_marker, extra_proc, _ = self.process_one_frame()
            if status==Status.Finished:
                break
            if status==Status.Skip:
                continue
            # store outputs
            for p in plane:
                poses_out[p].append(plane[p])
            for i in individual_marker:
                individual_markers_out[i].append(individual_marker[i])
            for e in extra_proc:
                extra_processing_out[e].append(extra_proc[e])

        return poses_out, individual_markers_out, extra_processing_out


# -*- coding: utf-8 -*-
# try if we have the dependencies for the GUI submodule and import it if so
try:
    import imgui_bundle
except ImportError:
    _has_GUI = False
else:
    _has_GUI = True
    from . import gui

from . import importing, validation
from .version import __version__, __url__, __author__, __email__, __description__


# ensure ffmpeg binaries needed by various submodules are on path
import ffmpeg as _ffmpeg
_ffmpeg.add_to_path()


import numpy as np
import cv2
import typing

from . import marker, ocv

M = typing.TypeVar("M", bound=int)
N = typing.TypeVar("N", bound=int)

def to_norm_pos(x,y,bbox):
    # transforms input (x,y) which is on a plane in world units
    # (e.g. mm on an aruco poster) to a normalized position
    # in an image of the plane, given the image's bounding box in
    # world units
    # for input (0,0) is bottom left, for output (0,0) is top left
    # bbox is [left, top, right, bottom]

    extents = [bbox[2]-bbox[0], bbox[1]-bbox[3]]
    pos     = [(x-bbox[0])/extents[0], (bbox[1]-y)/extents[1]]    # bbox[1]-y instead of y-bbox[3] to flip y
    return pos

def to_image_pos(x,y,bbox,img_size,margin=[0,0]):
    # transforms input (x,y) which is on a plane in world units
    # (e.g. mm on an aruco poster) to a pixel position in the
    # image, given the image's bounding box in world units
    # imSize should be active image area in pixels, excluding margin

    # fractional position between bounding box edges, (0,0) in bottom left
    pos = to_norm_pos(x,y, bbox)
    # turn into int, add margin
    pos = [p*s+m for p,s,m in zip(pos,img_size,margin)]
    return pos

def in_bbox(x,y,bbox,margin=0.):
    pos = to_norm_pos(x,y,bbox)
    return (pos[0]>=-margin and pos[0]<=1+margin) and (pos[1]>=-margin and pos[1]<=1+margin)

def dist_from_bbox(x,y,bbox):
    pos = to_norm_pos(x,y,bbox)
    if (pos[0]>=0 and pos[0]<=1) and (pos[1]>=0 and pos[1]<=1):
        return 0.   # inside bbox
    # compute max distance from edge of bbox
    dx = pos[0] if pos[0]<0. else pos[0]-1
    dy = pos[1] if pos[1]<0. else pos[1]-1
    return abs(max(dx,dy))


def estimate_homography_known_marker(known: list[marker.Marker], detected_corners, detected_IDs):
    # collect matching corners in image and in world
    img_points = []
    obj_points = []
    detected_IDs = detected_IDs.flatten()
    if len(detected_IDs) != len(detected_corners):
        raise ValueError('unexpected number of IDs (%d) given number of corner arrays (%d)' % (len(detected_IDs),len(detected_corners)))
    for i in range(0, len(detected_IDs)):
        if detected_IDs[i] in known:
            dc = detected_corners[i]
            if dc.shape[0]==1 and dc.shape[1]==4:
                dc = np.reshape(dc,(4,1,2))
            img_points.extend([x.flatten() for x in dc])
            obj_points.extend(known[detected_IDs[i]].corners)

    if len(img_points) < 4:
        return None

    # compute Homography
    return estimate_homography(obj_points, img_points)

def estimate_homography(obj_points, img_points):
    img_points = np.float32(img_points)
    obj_points = np.float32(obj_points)
    h, _ = cv2.findHomography(img_points, obj_points)
    return h

def apply_homography(points, H):
    if np.any(np.isnan(points)):
        return np.full_like(points, np.nan)

    return cv2.perspectiveTransform(points.astype('float').reshape((-1,1,2)),H).reshape((-1,2))


def distort_points(points_cam: np.ndarray[tuple[M, typing.Literal[2]], np.dtype[np.float64]], cam_params: ocv.CameraParams) -> np.ndarray[tuple[M, typing.Literal[2]], np.dtype[np.float64]]:
    if np.any(np.isnan(points_cam)):
        return np.full_like(points_cam, np.nan)

    if cam_params.has_colmap_camera():
        # unproject, ignoring distortion as this is an undistorted point
        points_w = cam_params.colmap_camera_no_distortion.cam_from_img(points_cam.reshape((-1,2)))
        # reproject, applying distortion
        return cam_params.colmap_camera.img_from_cam(points_w)
    elif cam_params.has_opencv_camera():
        # unproject, ignoring distortion as this is an undistorted point
        points_w = cv2.undistortPoints(points_cam.astype('float'), cam_params.camera_mtx, np.zeros((1, 5)))
        # reproject, applying distortion
        return cv2.projectPoints(cv2.convertPointsToHomogeneous(points_w), np.zeros((1, 1, 3)), np.zeros((1, 1, 3)), cam_params.camera_mtx, cam_params.distort_coeffs)[0].reshape((-1,2))
    else:
        return np.full_like(points_cam, np.nan)

def undistort_points(points_cam: np.ndarray[tuple[M, typing.Literal[2]], np.dtype[np.float64]], cam_params: ocv.CameraParams) -> np.ndarray[tuple[M, typing.Literal[2]], np.dtype[np.float64]]:
    if np.any(np.isnan(points_cam)):
        return np.full_like(points_cam, np.nan)

    if cam_params.has_colmap_camera():
        # unproject, removing distortion
        points_w = cam_params.colmap_camera.cam_from_img(points_cam.reshape((-1,2)))
        # reproject, without applying distortion
        return cam_params.colmap_camera_no_distortion.img_from_cam(points_w)
    elif cam_params.has_opencv_camera():
        return cv2.undistortPoints(points_cam.astype('float'), cam_params.camera_mtx, cam_params.distort_coeffs, P=cam_params.camera_mtx).reshape((-1,2)) # P=cameraMatrix to reproject to camera
    else:
        return np.full_like(points_cam, np.nan)

def unproject_points(points_cam: np.ndarray[tuple[M, typing.Literal[2]], np.dtype[np.float64]], cam_params: ocv.CameraParams) -> np.ndarray[tuple[M, typing.Literal[3]], np.dtype[np.float64]]:
    if np.any(np.isnan(points_cam)):
        return np.full((points_cam.shape[0],3), np.nan)

    if cam_params.has_colmap_camera():
        return cv2.convertPointsToHomogeneous(cam_params.colmap_camera.cam_from_img(points_cam.reshape((-1,2)))).reshape((-1,3))
    elif cam_params.has_opencv_camera():
        points_w = cv2.undistortPoints(points_cam.reshape((-1,2)).astype('float'), cam_params.camera_mtx, cam_params.distort_coeffs).reshape((-1,2))
        return cv2.convertPointsToHomogeneous(points_w).reshape((-1,3))
    else:
        return np.full((points_cam.shape[0],3), np.nan)

def project_points(points_world: np.ndarray[tuple[M, typing.Literal[3]], np.dtype[np.float64]], cam_params: ocv.CameraParams, ignore_distortion=False, rot_vec: np.ndarray[tuple[typing.Literal[3]], np.dtype[np.float64]]=None, trans_vec: np.ndarray[tuple[typing.Literal[3]], np.dtype[np.float64]]=None) -> np.ndarray[tuple[M, typing.Literal[2]], np.dtype[np.float64]]:
    if np.any(np.isnan(points_world)):
        return np.full((points_world.shape[0],2), np.nan)

    if cam_params.has_colmap_camera():
        if rot_vec is not None and trans_vec is not None:
            RMat = cv2.Rodrigues(rot_vec)[0]
            RtMat = np.hstack((RMat, trans_vec.reshape(3,1)))
            points_world = np.matmul(RtMat,cv2.convertPointsToHomogeneous(points_world.reshape((-1,3))).reshape((-1,4)).T).T
        if ignore_distortion:
            return cam_params.colmap_camera_no_distortion.img_from_cam(points_world.reshape((-1,3)))
        else:
            return cam_params.colmap_camera.img_from_cam(points_world.reshape((-1,3))).reshape((-1,2))
    elif cam_params.has_opencv_camera():
        return cv2.projectPoints(points_world.astype('float'),
                                 np.zeros((1, 1, 3)) if rot_vec is None else rot_vec,
                                 np.zeros((1, 1, 3)) if trans_vec is None else trans_vec,
                                 cam_params.camera_mtx,
                                 np.zeros((1, 5)) if ignore_distortion else cam_params.distort_coeffs)[0].reshape((-1,2))
    else:
        return np.full((points_world.shape[0],2), np.nan)


def intersect_plane_ray(plane_normal, plane_point, ray_direction, ray_point, epsilon=1e-6):
    # from https://rosettacode.org/wiki/Find_the_intersection_of_a_line_with_a_plane#Python

    ndotu = plane_normal.dot(ray_direction)
    if abs(ndotu) < epsilon:
        # raise RuntimeError("no intersection or line is within plane")
        return np.full((3,), np.nan)

    w = ray_point - plane_point
    si = -plane_normal.dot(w) / ndotu
    return w + si * ray_direction + plane_point


def angle_between(v1, v2):
    return (180.0 / np.pi) * np.arctan2(np.linalg.norm(np.cross(v1,v2)), np.dot(v1,v2))


import pathlib
import importlib.resources
import shutil

from ... import aruco


def deploy_config(output_dir: str|pathlib.Path, overwrite=False) -> list[str]:
    output_dir = pathlib.Path(output_dir)
    if not output_dir.is_dir():
        raise RuntimeError('the requested directory "%s" does not exist' % output_dir)

    # copy over all config files
    not_copied: list[str] = []
    for r in ['markerPositions.csv', 'targetPositions.csv', 'validationSetup.txt']:
        out_file = output_dir/r
        if out_file.exists() and not overwrite:
            not_copied.append(r)
            continue
        with importlib.resources.path(__package__, r) as p:
            shutil.copyfile(p, out_file)

    not_copied.extend(deploy_maker(output_dir))
    return not_copied

def deploy_maker(output_dir: str|pathlib.Path, overwrite=False) -> list[str]:
    output_dir = pathlib.Path(output_dir)
    if not output_dir.is_dir():
        raise RuntimeError('the requested directory "%s" does not exist' % output_dir)

    # copy over all files
    not_copied: list[str] = []
    for r in ['poster.tex']:
        out_file = output_dir/r
        if out_file.exists() and not overwrite:
            not_copied.append(r)
            continue
        with importlib.resources.path(__package__, r) as p:
            shutil.copyfile(p, out_file)

    deploy_marker_images(output_dir)    # N.B. these can be safely overwritten
    return not_copied

def deploy_marker_images(output_dir: str|pathlib.Path):
    from ..config import get_validation_setup

    output_dir = pathlib.Path(output_dir) / "all-markers"
    if not output_dir.is_dir():
        output_dir.mkdir()

    # get validation setup
    validationSetup = get_validation_setup()

    # generate and store the markers
    aruco.deploy_marker_images(output_dir, 1000, validationSetup['arucoDictionary'], validationSetup['markerBorderBits'])

def deploy_default_pdf(output_file_or_dir: str|pathlib.Path):
    output_file_or_dir = pathlib.Path(output_file_or_dir)
    if output_file_or_dir.is_dir():
        output_file_or_dir = output_file_or_dir / 'poster.pdf'

    with importlib.resources.path(__package__,'poster.pdf') as p:
        shutil.copyfile(p, str(output_file_or_dir))



import numpy as np
import cv2
import pathlib
from typing import Any, TypedDict

from . import drawing, marker, ocv, plane, pose

default_dict = cv2.aruco.DICT_4X4_250

dict_id_to_str: dict[int,str] = {getattr(cv2.aruco,k):k for k in ['DICT_4X4_50', 'DICT_4X4_100', 'DICT_4X4_250', 'DICT_4X4_1000', 'DICT_5X5_50', 'DICT_5X5_100', 'DICT_5X5_250', 'DICT_5X5_1000', 'DICT_6X6_50', 'DICT_6X6_100', 'DICT_6X6_250', 'DICT_6X6_1000', 'DICT_7X7_50', 'DICT_7X7_100', 'DICT_7X7_250', 'DICT_7X7_1000', 'DICT_ARUCO_ORIGINAL', 'DICT_APRILTAG_16H5', 'DICT_APRILTAG_25H9', 'DICT_APRILTAG_36H10', 'DICT_APRILTAG_36H11', 'DICT_ARUCO_MIP_36H12']}
def str_to_dict_id(aruco_dict_name: str):
    if not hasattr(cv2.aruco,aruco_dict_name):
        raise ValueError(f'ArUco dictionary with name "{aruco_dict_name}" is not known.')
    return getattr(cv2.aruco,aruco_dict_name)

# same number means that the dictionaries are the same (i.e. marker 3 is the same for all the dictionaries of the same family), just different number of markers in the dictionary
dict_id_to_family = {
    cv2.aruco.DICT_4X4_50:          0,
    cv2.aruco.DICT_4X4_100:         0,
    cv2.aruco.DICT_4X4_250:         0,
    cv2.aruco.DICT_4X4_1000:        0,
    cv2.aruco.DICT_5X5_50:          1,
    cv2.aruco.DICT_5X5_100:         1,
    cv2.aruco.DICT_5X5_250:         1,
    cv2.aruco.DICT_5X5_1000:        1,
    cv2.aruco.DICT_6X6_50:          2,
    cv2.aruco.DICT_6X6_100:         2,
    cv2.aruco.DICT_6X6_250:         2,
    cv2.aruco.DICT_6X6_1000:        2,
    cv2.aruco.DICT_7X7_50:          3,
    cv2.aruco.DICT_7X7_100:         3,
    cv2.aruco.DICT_7X7_250:         3,
    cv2.aruco.DICT_7X7_1000:        3,
    cv2.aruco.DICT_ARUCO_ORIGINAL:  4,
    cv2.aruco.DICT_APRILTAG_16H5:   5,
    cv2.aruco.DICT_APRILTAG_25H9:   6,
    cv2.aruco.DICT_APRILTAG_36H10:  7,
    cv2.aruco.DICT_APRILTAG_36H11:  8,
    cv2.aruco.DICT_ARUCO_MIP_36H12: 9
}
family_to_str = {
    0: ('DICT_4X4', True),
    1: ('DICT_5X5', True),
    2: ('DICT_6X6', True),
    3: ('DICT_7X7', True),
    4: ('DICT_ARUCO_ORIGINAL', False),
    5: ('DICT_APRILTAG_16H5', False),
    6: ('DICT_APRILTAG_25H9', False),
    7: ('DICT_APRILTAG_36H10', False),
    8: ('DICT_APRILTAG_36H11', False),
    9: ('DICT_ARUCO_MIP_36H12', False),
}

class PlaneSetup(TypedDict):
    plane                   : plane.Plane
    aruco_detector_params   : dict[str,Any]
    aruco_refine_params     : dict[str,Any]
    min_num_markers         : int
class MarkerSetup(TypedDict):
    aruco_detector_params   : dict[str,Any]
    detect_only             : bool
    size                    : float

def reduce_to_families(dictionary_ids: list[int]) -> tuple[list[int],dict[int,int]]:
    # turn into families, and the largest dict necessary per family
    # get unique dictionaries
    seen: set[int] = set()
    aruco_dicts = [x for x in dictionary_ids if x not in seen and not seen.add(x)]
    # first organize by family
    by_family: dict[int,list[int]] = {}
    for d in aruco_dicts:
        f = dict_id_to_family[d]
        if f not in by_family:
            by_family[f] = []
        by_family[f].append(d)
    # for each family, if there are multiple dicts, get the largest
    needed_dicts = [sorted(by_family[f], key=lambda x: get_dict_size(x))[-1] for f in by_family]
    # make a mapping of dictionary (requested) to dictionary (used)
    aruco_dict_mapping = {d:d2 for f,d2 in zip(by_family,needed_dicts) for d in by_family[f]}
    return needed_dicts, aruco_dict_mapping

def get_dict_size(dictionary_id: int) -> int:
    return cv2.aruco.getPredefinedDictionary(dictionary_id).bytesList.shape[0]

def get_marker_image(size: int, m_id: int, ArUco_dict_id: int, marker_border_bits: int) -> np.ndarray|None:
    if m_id>=get_dict_size(ArUco_dict_id):
        return None
    marker_image = np.zeros((size, size), dtype=np.uint8)
    return cv2.aruco.generateImageMarker(cv2.aruco.getPredefinedDictionary(ArUco_dict_id), m_id, size, marker_image, marker_border_bits)

def deploy_marker_images(output_dir: str|pathlib.Path, size: int, ArUco_dict_id: int, marker_border_bits: int=1):
    # Generate the markers
    for m_id in range(get_dict_size(ArUco_dict_id)):
        marker_image = get_marker_image(size, m_id, ArUco_dict_id, marker_border_bits)
        if marker_image is not None:
            cv2.imwrite(output_dir / f"{m_id}.png", marker_image)

class Detector:
    def __init__(self, dictionary_id: int):
        self.dictionary_id  = dictionary_id
        self._family        = dict_id_to_family[self.dictionary_id]
        self._is_family     = family_to_str[self._family][1]

        self.planes             : dict[str, PlaneSetup]         = {}
        self._boards            : dict[str, cv2.aruco.Board]    = {}
        self.individual_markers : dict[int, MarkerSetup]        = {}
        self._indiv_marker_points:dict[int, np.ndarray]         = {}

        self._plane_marker_ids      : dict[str,set[int]]        = {}
        self._individual_marker_ids : set[int]                  = set()
        self._all_markers           : set[int]                  = set()

        self._user_detector_params  : dict[str]                 = {}
        self._user_refine_params    : dict[str]                 = {}

        self._det: cv2.aruco.ArucoDetector                      = None

        self._last_detect_output : tuple[dict[str,dict[str]],dict[str],dict[str],list[np.ndarray]] = {}

    def add_plane(self, name: str, setup: PlaneSetup):
        self._check_dict(setup['plane'].aruco_dict_id, 'plane')
        self.planes[name] = setup
        if 'aruco_detector_params' in self.planes[name] and self.planes[name]['aruco_detector_params']:
            self._update_parameters('detector', self.planes[name]['aruco_detector_params'])
        if 'aruco_refine_params' in self.planes[name] and self.planes[name]['aruco_refine_params']:
            self._update_parameters('refine'  , self.planes[name]['aruco_refine_params'])
        self._boards[name]= self.planes[name]['plane'].get_aruco_board()

        markers = self.planes[name]['plane'].get_marker_IDs()
        for ms in markers:
            if ms!='plane':
                continue
            m_ids = {m.m_id for m in markers[ms]}
            self._all_markers.update(m_ids)
            self._plane_marker_ids[name] = m_ids

    def add_individual_marker(self, mark: marker.MarkerID, setup: MarkerSetup):
        self._check_dict(mark.aruco_dict_id, 'individual marker')
        self.individual_markers[mark.m_id] = setup
        if 'aruco_detector_params' in self.individual_markers[mark.m_id] and self.individual_markers[mark.m_id]['aruco_detector_params']:
            self._update_parameters('detector', self.individual_markers[mark.m_id]['aruco_detector_params'])
        self._all_markers.add(mark.m_id)
        self._individual_marker_ids.add(mark.m_id)
        # get marker points in world
        marker_size = self.individual_markers[mark.m_id].get('size',None) if not self.individual_markers[mark.m_id].get('detect_only',False) else None
        if not marker_size or marker_size<0.:
            marker_points = None
        else:
            marker_points =  np.array([[-marker_size/2,  marker_size/2, 0],
                                       [ marker_size/2,  marker_size/2, 0],
                                       [ marker_size/2, -marker_size/2, 0],
                                       [-marker_size/2, -marker_size/2, 0]])
        self._indiv_marker_points[mark.m_id] = marker_points

    def _check_dict(self, dict_id: int, what: str):
        if self._is_family:
            family = dict_id_to_family[dict_id]
            if family!=self._family:
                raise ValueError(f'The dictionary for this new {what}, {dict_id_to_str[dict_id]}, is not part of the family ({family_to_str[family][0]}) used for this detector. Use dictionary {dict_id_to_str[self.dictionary_id]} or smaller.')
            elif dict_id>self.dictionary_id:
                raise ValueError(f'The dictionary for this new {what}, {dict_id_to_str[dict_id]}, contains more markers than the dictionary used for this detector ({dict_id_to_str[self.dictionary_id]}). Use a dictionary with more markers when creating this detector.')
        elif dict_id!=self.dictionary_id:
            raise ValueError(f'The dictionary for this new {what}, {dict_id_to_str[dict_id]}, does not match the dictionary used for this detector ({dict_id_to_str[self.dictionary_id]}).')

    def _update_parameters(self, which: str, new_params: dict):
        if which=='detector':
            param_dict = self._user_detector_params
            cls = cv2.aruco.DetectorParameters
        elif 'refine':
            param_dict = self._user_refine_params
            cls = cv2.aruco.RefineParameters
        else:
            raise ValueError(f'parameter type "{which}" not understood')
        for p in new_params:
            if not hasattr(cls, p):
                raise AttributeError(f'{p} is not a valid parameter for cv2.aruco.{cls.__name__}')
            if p in param_dict and new_params[p]!=param_dict[p]:
                fam_str,is_family = family_to_str[dict_id_to_family[self.dictionary_id]]
                dict_str = f'{fam_str} family' if is_family else f'{dict_id_to_str[self.dictionary_id]} dictionary'
                raise ValueError(f'You have already set the parameter {p} to {param_dict[p]} and are now trying to set it to {new_params[p]}, in the detector for the {dict_str}. Resolve this conflict by checking this setting for all planes and individual markers using the {dict_str}.')
            param_dict[p] = new_params[p]

    def create_detector(self):
        # set detector parameters
        detector_params                       = cv2.aruco.DetectorParameters()
        detector_params.cornerRefinementMethod= cv2.aruco.CORNER_REFINE_SUBPIX    # good default, user can override
        refine_params                         = cv2.aruco.RefineParameters()
        for p in self._user_detector_params:
            setattr(detector_params, p, self._user_detector_params[p])
        for p in self._user_refine_params:
            setattr(refine_params, p, self._user_refine_params[p])

        self._det = cv2.aruco.ArucoDetector(cv2.aruco.getPredefinedDictionary(self.dictionary_id),
                                            detector_params, refine_params)

    def detect_markers(self, image: cv2.UMat, camera_params: ocv.CameraParams) -> tuple[dict[str,dict[str]],dict[str],dict[str],list[np.ndarray]]:
        img_points, ids, rejected_img_points = self._detect_markers(image, self._det)

        # For each plane, refine detected markers (eliminates markers not part of the plane, adds missing markers to the poster)
        out_planes: dict[str] = {}
        for p in self.planes:
            if ids is not None and len(ids)>self.planes[p]['min_num_markers']:
                img_points, ids, rejected_img_points, recovered_ids = \
                    self._refine_detection(image, img_points, ids, rejected_img_points,
                                           self._det, self._boards[p],
                                           camera_params.camera_mtx, camera_params.distort_coeffs)
                out_planes[p] = dict(zip(['img_points', 'ids', 'recovered_ids'],(img_points, ids, recovered_ids)))
            else:
                out_planes[p] = None if ids is None else dict(zip(['img_points', 'ids', 'recovered_ids'],(img_points, ids, None)))
            if out_planes[p]:
                out_planes[p]['img_points'],out_planes[p]['ids'] = self._filter_detections(out_planes[p]['img_points'], out_planes[p]['ids'], self._plane_marker_ids[p])

        # For individual markers, only keep the ones known
        out_individual: dict[str] = {}
        out_individual['img_points'],out_individual['ids'] = self._filter_detections(img_points, ids, self._individual_marker_ids)

        # collect unexpected markers
        unexpected_markers: dict[str] = {}
        unexpected_markers['img_points'],unexpected_markers['ids'] = self._filter_detections(img_points, ids, self._all_markers, keep_expected=False)

        self._last_detect_output = (out_planes, out_individual, unexpected_markers, rejected_img_points)
        return self._last_detect_output

    def _detect_markers(self, image: cv2.UMat, det: cv2.aruco.ArucoDetector):
        img_points, ids, rejected_img_points = det.detectMarkers(image)
        if np.any(ids==None):
            ids = None
        return img_points, ids, rejected_img_points

    def _refine_detection(self, image: cv2.UMat, detected_corners, detected_ids, rejected_corners, det: cv2.aruco.ArucoDetector, board: cv2.aruco.Board, camera_mtx, distort_coeffs):
        img_points, ids, rejected_img_points, recovered_ids = \
            det.refineDetectedMarkers(
                image = image, board = board,
                detectedCorners = detected_corners, detectedIds = detected_ids, rejectedCorners = rejected_corners,
                cameraMatrix = camera_mtx, distCoeffs = distort_coeffs
                )
        if img_points and img_points[0].shape[0]==4:
            # there are versions out there where there is a bug in output shape of each set of corners, fix up
            img_points = [np.reshape(c,(1,4,2)) for c in img_points]
        if rejected_img_points and rejected_img_points[0].shape[0]==4:
            # same as for corners
            rejected_img_points = [np.reshape(c,(1,4,2)) for c in rejected_img_points]

        return img_points, ids, rejected_img_points, recovered_ids

    def _filter_detections(self, img_points: list[np.ndarray], ids: np.ndarray, expected_ids: list[np.ndarray], keep_expected=True):
        if ids is None or not img_points:
            return img_points, ids
        if not keep_expected:
            expected_ids = set(ids.flatten())-set(expected_ids)
        if not expected_ids:
            # optimization. If output will definitely be empty, return directly
            return tuple(), None
        to_remove = np.where([x not in expected_ids for x in ids.flatten()])[0]
        ids = np.delete(ids, to_remove, axis=0)
        img_points = tuple(v for i,v in enumerate(img_points) if i not in to_remove)
        return img_points, ids

    def get_matching_image_board_points(self, plane_name: str, detect_tuple=None):
        if detect_tuple is None:
            detect_tuple = self._last_detect_output
        if plane_name not in detect_tuple[0] or detect_tuple[0][plane_name]['ids'] is None or not detect_tuple[0][plane_name]['img_points']:
            return None, None
        objP, imgP = self._boards[plane_name].matchImagePoints(detect_tuple[0][plane_name]['img_points'], detect_tuple[0][plane_name]['ids'])
        if imgP is None or int(imgP.shape[0]/4)<self.planes[plane_name]['min_num_markers']:
            return None, None
        return imgP, objP

    def get_individual_marker_points(self, marker_id: int, detect_tuple=None):
        if detect_tuple is None:
            detect_tuple = self._last_detect_output
        if detect_tuple[1]['ids'] is None or not detect_tuple[1]['img_points'] or marker_id not in detect_tuple[1]['ids']:
            return None, None
        img_points = detect_tuple[1]['img_points'][detect_tuple[1]['ids'].flatten().tolist().index(marker_id)]
        return img_points, self._indiv_marker_points[marker_id]

    def visualize(self, frame, detect_tuple=None, sub_pixel_fac=8, plane_marker_color=(0,255,0), recovered_plane_marker_color=(255,255,0), individual_marker_color=(255,0,255), unexpected_marker_color=(150,253,253), rejected_marker_color=None):
        if detect_tuple is None:
            detect_tuple = self._last_detect_output
        special_highlight = []

        # for debug, can draw rejected markers on frame
        if rejected_marker_color is not None:
            cv2.aruco.drawDetectedMarkers(frame, detect_tuple[3], None, borderColor=rejected_marker_color)

        # draw detected markers on the frame
        if plane_marker_color is not None:
            for p in detect_tuple[0]:
                if not detect_tuple[0][p] or 'ids' not in detect_tuple[0][p] or len(detect_tuple[0][p]['ids'])==0:
                    continue
                if recovered_plane_marker_color is not None and detect_tuple[0][p]['recovered_ids'] is not None and len(detect_tuple[0][p]['recovered_ids'])>0:
                    special_highlight = [detect_tuple[0][p]['recovered_ids'],recovered_plane_marker_color]
                drawing.arucoDetectedMarkers(frame, detect_tuple[0][p]['img_points'], detect_tuple[0][p]['ids'], border_color=plane_marker_color, sub_pixel_fac=sub_pixel_fac, special_highlight=special_highlight)
        if individual_marker_color is not None and detect_tuple[1]['ids'] is not None and len(detect_tuple[1]['ids']>0):
            drawing.arucoDetectedMarkers(frame, detect_tuple[1]['img_points'], detect_tuple[1]['ids'], border_color=individual_marker_color, sub_pixel_fac=sub_pixel_fac)
        if unexpected_marker_color is not None and detect_tuple[2]['ids'] is not None and len(detect_tuple[2]['ids']>0):
            drawing.arucoDetectedMarkers(frame, detect_tuple[2]['img_points'], detect_tuple[2]['ids'], border_color=unexpected_marker_color, sub_pixel_fac=sub_pixel_fac)

class Manager:
    # takes single planes and individual markers, and for each information about when they should be
    # detected, and consolidates them into a minimal set of detectors
    # with all planes/individual markers associated to one of these detectors
    def __init__(self):
        # planes to be detected
        self.planes                 : dict[str, PlaneSetup]                 = {}
        self.plane_proc_intervals   : dict[str, list[int]|list[list[int]]]  = {}
        self._plane_to_detector     : dict[str, int]                        = {}
        # individual markers to be detected
        self.individual_markers                 : dict[marker.MarkerID, MarkerSetup]    = {}
        self.individual_markers_proc_intervals  : dict[str, list[int]|list[list[int]]]  = {}

        # consolidated into set of detectors, and associated planes+individual markers for each
        self._detectors             : dict[int, Detector]                   = {}
        self._det_cache             : dict[int, tuple[int,tuple]]           = {}
        self._last_viz_frame_idx    : dict[int,int]                         = {}

        # colors for drawing (in BGR order)
        self._plane_marker_color            = (  0,255,  0)
        self._recovered_plane_marker_color  = (255,255,  0)
        self._individual_marker_color       = (255,  0,255)
        self._unexpected_marker_color       = (128,255,255)
        self._rejected_marker_color         = None # by default not drawn. If wanted, (0,0,255) is a good color

    def add_plane(self, plane: str, planes_setup: PlaneSetup, processing_intervals: list[int]|list[list[int]] = None):
        if plane in self.planes:
            raise ValueError(f'Cannot register the plane "{plane}", it is already registered')
        self.planes[plane]                  = planes_setup
        self.plane_proc_intervals[plane]    = processing_intervals

    def add_individual_marker(self, mark: marker.MarkerID, marker_setup: MarkerSetup, processing_intervals: list[int]|list[list[int]] = None):
        if mark in self.individual_markers:
            raise ValueError(f'Cannot register the individual marker {marker.marker_ID_to_str(mark)}, it is already registered')
        self.individual_markers[mark]                = marker_setup
        self.individual_markers_proc_intervals[mark] = processing_intervals

    def set_visualization_colors(self, plane_marker_color=(0,255,0), recovered_plane_marker_color=(0,255,255), individual_marker_color=(255,0,255), unexpected_marker_color=(255,255,128), rejected_marker_color=None):
        # user should provide colors in RGB, internally we store as BGR
        if plane_marker_color is not None:
            plane_marker_color = plane_marker_color[::-1]
        self._plane_marker_color = plane_marker_color
        if recovered_plane_marker_color is not None:
            recovered_plane_marker_color = recovered_plane_marker_color[::-1]
        self._recovered_plane_marker_color = recovered_plane_marker_color
        if individual_marker_color is not None:
            individual_marker_color = individual_marker_color[::-1]
        self._individual_marker_color = individual_marker_color
        if unexpected_marker_color is not None:
            unexpected_marker_color = unexpected_marker_color[::-1]
        self._unexpected_marker_color = unexpected_marker_color
        if rejected_marker_color is not None:
            rejected_marker_color = rejected_marker_color[::-1]
        self._rejected_marker_color = rejected_marker_color

    def consolidate_setup(self):
        # get list of all ArUco dicts and markers we're dealing with
        all_markers: set[marker.MarkerID] = set()
        for p in self.planes:
            markers = self.planes[p]['plane'].get_marker_IDs()
            for ms in markers:
                if ms!='plane':
                    # N.B.: other markers should be registered by caller as individual markers
                    continue
                if all_markers.intersection(markers[ms]):
                    raise RuntimeError('Markers are not unique')
                all_markers.update(markers[ms])
        for m in self.individual_markers:
            if m in all_markers:
                raise RuntimeError('Markers are not unique')
            all_markers.add(m)

        # see for which marker dicts we need detectors to service all these
        # also determine mapping of requested ArUco dicts to these detectors
        needed_dicts, dict_mapping = reduce_to_families({m.aruco_dict_id for m in all_markers})

        # organize planes and individual markers into the dict that will be used for their detection
        planes_organized        : dict[int,list[str]]             = {d:[] for d in needed_dicts}
        indiv_markers_organized : dict[int,list[marker.MarkerID]] = {d:[] for d in needed_dicts}
        for p in self.planes:
            det_dict = dict_mapping[self.planes[p]['plane'].aruco_dict_id]
            planes_organized[det_dict].append(p)
        for m in self.individual_markers:
            det_dict = dict_mapping[m.aruco_dict_id]
            indiv_markers_organized[det_dict].append(m)

        # make the needed detectors
        self._detectors.clear()
        self._plane_to_detector.clear()
        for d in needed_dicts:
            self._detectors[d] = Detector(d)
            for p in planes_organized[d]:
                self._detectors[d].add_plane(p, self.planes[p])
                self._plane_to_detector[p] = d
            for m in indiv_markers_organized[d]:
                self._detectors[d].add_individual_marker(m, self.individual_markers[m])
            self._detectors[d].create_detector()

    def register_with_estimator(self, estimator: pose.Estimator):
        # this handles registration of all planes and individual markers with the estimator
        # and makes sure our wrapper function for the aruco detector gets called which handles
        # aruco detection so that each detector is run only once on a frame
        for p in self.planes:
            estimator.add_plane(p,
                                lambda pn, fi, fr, cp: self._detect_plane(pn, fi, fr, cp),
                                self.plane_proc_intervals[p],
                                lambda pn, fi, fr, _: self._visualize_plane(pn, fi, fr))
        for m in self.individual_markers:
            estimator.add_individual_marker(m,
                                            lambda k, fi, fr, cp: self._detect_individual_marker(k, fi, fr, cp),
                                            self.individual_markers_proc_intervals[m],
                                            lambda k, fi, fr, _: self._visualize_individual_marker(k, fi, fr))

    def _detect_plane(self, plane_name: str, frame_idx: int, frame: np.ndarray, camera_parameters: ocv.CameraParams):
        if plane_name not in self._plane_to_detector:
            raise ValueError(f'The plane {plane_name} is not known')
        aruco_dict_id = self._plane_to_detector[plane_name]
        detect_tuple = self._get_detector_cache(aruco_dict_id, frame_idx, frame, camera_parameters)
        if not detect_tuple[0] or plane_name not in detect_tuple[0] or not detect_tuple[0][plane_name]:
            return None, None
        return self._detectors[aruco_dict_id].get_matching_image_board_points(plane_name, detect_tuple)

    def _detect_individual_marker(self, mark: marker.MarkerID, frame_idx: int, frame: np.ndarray, camera_parameters: ocv.CameraParams):
        if mark not in self.individual_markers:
            raise ValueError(f'The individual marker {marker.marker_ID_to_str(mark)} is not known')
        detect_tuple = self._get_detector_cache(mark.aruco_dict_id, frame_idx, frame, camera_parameters)
        if not detect_tuple[1] or detect_tuple[1]['ids'] is None or mark.m_id not in detect_tuple[1]['ids']:
            return None, None
        return self._detectors[mark.aruco_dict_id].get_individual_marker_points(mark.m_id, detect_tuple)

    def _get_detector_cache(self, aruco_dict_id: int, frame_idx: int, frame: np.ndarray, camera_parameters: ocv.CameraParams):
        if aruco_dict_id not in self._det_cache or self._det_cache[aruco_dict_id][0]!=frame_idx:
            if frame is None:
                return None
            detect_tuple = self._detectors[aruco_dict_id].detect_markers(frame, camera_parameters)
            self._det_cache[aruco_dict_id] = (frame_idx, detect_tuple)
        return self._det_cache[aruco_dict_id][1]

    def _visualize_plane(self, plane_name: str, frame_idx: int, frame: np.ndarray):
        if plane_name not in self._plane_to_detector:
            raise ValueError(f'The plane {plane_name} is not known')
        aruco_dict_id = self._plane_to_detector[plane_name]
        if aruco_dict_id in self._last_viz_frame_idx and self._last_viz_frame_idx[aruco_dict_id]==frame_idx:
            # nothing to do, already drawn
            return
        detect_tuple = self._get_detector_cache(aruco_dict_id, frame_idx, None, None)
        if detect_tuple is not None:
            frame = self._detectors[aruco_dict_id].visualize(frame, detect_tuple, plane_marker_color=self._plane_marker_color, recovered_plane_marker_color=self._recovered_plane_marker_color, individual_marker_color=self._individual_marker_color, unexpected_marker_color=self._unexpected_marker_color, rejected_marker_color=self._rejected_marker_color)
            self._last_viz_frame_idx[aruco_dict_id] = frame_idx

    def _visualize_individual_marker(self, mark: marker.MarkerID, frame_idx: int, frame: np.ndarray):
        if mark not in self.individual_markers:
            raise ValueError(f'The individual marker {marker.marker_ID_to_str(mark)} is not known')
        if mark.aruco_dict_id in self._last_viz_frame_idx and self._last_viz_frame_idx[mark.aruco_dict_id]==frame_idx:
            # nothing to do, already drawn
            return
        detect_tuple = self._get_detector_cache(mark.aruco_dict_id, frame_idx, None, None)
        if detect_tuple is not None:
            frame = self._detectors[mark.aruco_dict_id].visualize(frame, detect_tuple, plane_marker_color=self._plane_marker_color, recovered_plane_marker_color=self._recovered_plane_marker_color, individual_marker_color=self._individual_marker_color, unexpected_marker_color=self._unexpected_marker_color, rejected_marker_color=self._rejected_marker_color)
            self._last_viz_frame_idx[mark.aruco_dict_id] = frame_idx


def create_board(board_corner_points: list[np.ndarray], ids: list[int], ArUco_dict: cv2.aruco.Dictionary):
    board_corner_points = np.dstack(board_corner_points)        # list of 2D arrays -> 3D array
    board_corner_points = np.rollaxis(board_corner_points,-1)   # 4x2xN -> Nx4x2
    board_corner_points = np.pad(board_corner_points,((0,0),(0,0),(0,1)),'constant', constant_values=(0.,0.)) # Nx4x2 -> Nx4x3 (at Z=0 to all points)
    return cv2.aruco.Board(board_corner_points, ArUco_dict, np.array(ids))


import math
import numpy as np
import pandas as pd
import typing
import pathlib

import I2MC
import matplotlib.pyplot as plt

from . import gaze_worldref, naming

# run I2MC on data projected to a plane
def from_plane_gaze(
        gazes: str|pathlib.Path|dict[int, list[gaze_worldref.Gaze]],
        classification_intervals: list[list[int]],
        output_directory: str|pathlib.Path,
        I2MC_settings_override: dict[str,typing.Any]=None,
        filename_stem: str = naming.fixation_classification_prefix,
        do_plot = True,
        plot_limits: list[list[float]] = None
    ):
    output_directory = pathlib.Path(output_directory)

    # read input if needed
    if not isinstance(gazes,dict):
        gazes = gaze_worldref.read_dict_from_file(gazes)

    # set I2MC options
    opt = {'xres': None, 'yres': None}  # dummy values for required options
    opt['missingx']         = math.nan
    opt['missingy']         = math.nan
    opt['maxdisp']          = 50        # mm
    opt['windowtimeInterp'] = .25       # s
    opt['maxMergeDist']     = 20        # mm
    opt['maxMergeTime']     = 81        # ms
    opt['minFixDur']        = 50        # ms

    # decide what sampling frequency to tell I2MC about. It doesn't work with varying sampling frequency, nor
    # any random sampling frequency. For our purposes, getting it right is not important (internally I2MC only
    # uses sampling frequency for converting some of the time units to samples, other things are taken directly
    # from the time signal). So, we have working I2MC settings for a few sampling frequencies, and just choose
    # the nearest based on empirically determined sampling frequency.
    ts          = np.array([s.timestamp for v in gazes.values() for s in v])
    recFreq     = np.round(np.mean(1000./np.diff(ts)))    # Hz
    knownFreqs  = [30., 50., 60., 90., 120., 200.]
    opt['freq'] = knownFreqs[np.abs(knownFreqs - recFreq).argmin()]
    if opt['freq']==200.:
        pass    # defaults are good
    elif opt['freq']==120.:
        opt['downsamples']      = [2, 3, 5]
        opt['chebyOrder']       = 7
    elif opt['freq'] in [50., 60.]:
        opt['downsamples']      = [2, 5]
        opt['downsampFilter']   = False
    else:
        # 90 Hz, 30 Hz
        opt['downsamples']      = [2, 3]
        opt['downsampFilter']   = False

    # apply setting overrides from caller, if any
    if I2MC_settings_override:
        for k in I2MC_settings_override:
            if I2MC_settings_override[k] is not None:
                opt[k] = I2MC_settings_override[k]

    # collect data
    qHasLeft        = np.any(np.logical_not(np.isnan([s.gazePosPlane2DLeft               for v in gazes.values() for s in v])))
    qHasRight       = np.any(np.logical_not(np.isnan([s.gazePosPlane2DRight              for v in gazes.values() for s in v])))
    qHasRay         = np.any(np.logical_not(np.isnan([s.gazePosPlane2D_vidPos_ray        for v in gazes.values() for s in v])))
    qHasHomography  = np.any(np.logical_not(np.isnan([s.gazePosPlane2D_vidPos_homography for v in gazes.values() for s in v])))
    for idx,iv in enumerate(classification_intervals):
        gazesToClassify = {k:v for (k,v) in gazes.items() if k>=iv[0] and (iv[1]==-1 or k<=iv[1])}
        # Doing detection on the world data if available is good, but we should plot using the ray (if
        # available) or homography data, as that corresponds to the gaze visualization provided in the
        # software, and for some recordings/devices the world-based coordinates can be far off.
        if qHasRay:
            ray_x  = np.array([s.gazePosPlane2D_vidPos_ray[0] for v in gazesToClassify.values() for s in v])
            ray_y  = np.array([s.gazePosPlane2D_vidPos_ray[1] for v in gazesToClassify.values() for s in v])
        elif qHasHomography:
            homography_x  = np.array([s.gazePosPlane2D_vidPos_homography[0] for v in gazesToClassify.values() for s in v])
            homography_y  = np.array([s.gazePosPlane2D_vidPos_homography[1] for v in gazesToClassify.values() for s in v])

        data = {}
        data['time'] = np.array([s.timestamp for v in gazesToClassify.values() for s in v])
        qNeedRecalcFix = False
        if qHasLeft and qHasRight:
            # prefer using separate left and right eye signals, if available. Better I2MC robustness
            data['L_X']  = np.array([s.gazePosPlane2DLeft[0]  for v in gazesToClassify.values() for s in v])
            data['L_Y']  = np.array([s.gazePosPlane2DLeft[1]  for v in gazesToClassify.values() for s in v])
            data['R_X']  = np.array([s.gazePosPlane2DRight[0] for v in gazesToClassify.values() for s in v])
            data['R_Y']  = np.array([s.gazePosPlane2DRight[1] for v in gazesToClassify.values() for s in v])
            qNeedRecalcFix = True
        elif qHasRay:
            data['average_X']  = ray_x
            data['average_Y']  = ray_y
        elif qHasHomography:
            data['average_X']  = homography_x
            data['average_Y']  = homography_y
        else:
            raise RuntimeError('No data available to process')

        # run event classification to find fixations
        fixations, data_I2MC, par_I2MC = I2MC.I2MC(data, opt, False)

        # replace gaze data used for classification with gaze position on scene video
        # see note above for why
        if qNeedRecalcFix:
            data_I2MC = data_I2MC.drop(columns=['L_X','L_Y','R_X','R_Y'],errors='ignore')
            data_I2MC['average_X'] = ray_x if qHasRay else homography_x
            data_I2MC['average_Y'] = ray_y if qHasRay else homography_y
            # recalculate fixation positions based on gaze position on video data
            fixations = I2MC.get_fixations(data_I2MC['finalweights'].array, data_I2MC['time'].array, data_I2MC['average_X'], data_I2MC['average_Y'], data_I2MC['average_missing'], par_I2MC)

        # store to file
        fix_df = pd.DataFrame(fixations)
        fix_df.to_csv(output_directory / f'{filename_stem}_interval_{idx+1:02d}.tsv',
                      mode='w', na_rep='nan', sep='\t', index=False, float_format='%.3f')

        # make timeseries plot of gaze data with fixations
        if do_plot:
            f = I2MC.plot.data_and_fixations(data_I2MC, fixations, fix_as_line=True, unit='mm', res=plot_limits)
            plt.gca().invert_yaxis()
            f.savefig(str(output_directory / f'{filename_stem}_interval_{idx+1:02d}.png'))
            plt.close(f)


import numpy as np
import pandas as pd
import pathlib
import typing
from collections import defaultdict

from . import data_files, drawing, json, naming, ocv

class MarkerID(typing.NamedTuple):
    m_id:           int
    aruco_dict_id:  int
def marker_ID_to_str(m: MarkerID):
    from . import aruco
    return f'{m.m_id} ({aruco.dict_id_to_str[m.aruco_dict_id]})'
def _serialize_marker_id(m: MarkerID):
    from . import aruco
    return {'m_id':m.m_id, 'aruco_dict':aruco.dict_id_to_str[m.aruco_dict_id]}
def _deserialize_marker_id(m: dict[str,str|int]):
    from . import aruco
    return MarkerID(m_id = m['m_id'],
                    aruco_dict_id = aruco.str_to_dict_id(m['aruco_dict_id' if 'aruco_dict_id' in m else 'aruco_dict']))
json.register_type(json.TypeEntry(MarkerID, '__config.MarkerID__', _serialize_marker_id, _deserialize_marker_id))

class Marker:
    def __init__(self, key: int, center: np.ndarray, corners: list[np.ndarray]=None, color: str=None, rot: float=0.):
        self.key = key
        self.center = center
        self.corners = corners
        self.color = color
        self.rot = rot

    def __str__(self):
        ret = '[%d]: center @ (%.2f, %.2f), rot %.0f deg' % (self.key, self.center[0], self.center[1], self.rot)
        return ret

    def shift(self, offset=np.ndarray):
        self.center += offset
        if self.corners:
            for c in self.corners:
                c += offset

def corners_intersection(corners):
    line1 = ( corners[0], corners[2] )
    line2 = ( corners[1], corners[3] )
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return np.array( [x,y] ).astype('float32')

class Pose:
    # description of tsv file used for storage
    _columns_compressed = {'frame_idx': 1, 'R_vec': 3, 'T_vec': 3}
    _non_float          = {'frame_idx': int}

    def __init__(self,
                 frame_idx  : int,
                 R_vec      : np.ndarray= None,
                 T_vec      : np.ndarray= None):
        self.frame_idx  : int         = frame_idx
        self.R_vec      : np.ndarray  = R_vec
        self.T_vec      : np.ndarray  = T_vec

    def pose_successful(self):
        return self.R_vec is not None and self.T_vec is not None

    def draw_frame_axis(self, frame, camera_params: ocv.CameraParams, arm_length, sub_pixel_fac = 8):
        if not camera_params.has_intrinsics():
            return
        drawing.openCVFrameAxis(frame, camera_params, self.R_vec, self.T_vec, arm_length, 3, sub_pixel_fac)


def read_dict_from_file(fileName:str|pathlib.Path, episodes:list[list[int]]=None) -> dict[int,Pose]:
    return data_files.read_file(fileName,
                                Pose, True, True, False, False,
                                episodes=episodes)[0]

def write_list_to_file(poses: list[Pose], fileName:str|pathlib.Path, skip_failed=False):
    data_files.write_array_to_file(poses, fileName,
                                   Pose._columns_compressed,
                                   skip_all_nan=skip_failed)

def get_file_name(marker_id: int, aruco_dict_id: int, folder: str|pathlib.Path|None) -> pathlib.Path:
    from . import aruco
    file_name = f'{naming.marker_pose_prefix}{aruco.dict_id_to_str[aruco_dict_id]}_{marker_id}.tsv'
    if folder is None:
        return file_name
    folder = pathlib.Path(folder)
    return folder / file_name

def read_dataframe_from_file(marker_id: int, aruco_dict_id: int, folder: str|pathlib.Path) -> pd.DataFrame:
    file = get_file_name(marker_id, aruco_dict_id, folder)
    return pd.read_csv(file,sep='\t', dtype=defaultdict(lambda: float, **Pose._non_float))


@typing.overload
def code_for_presence(markers: pd.DataFrame, allow_failed=False) -> pd.DataFrame: ...
def code_for_presence(markers: dict[typing.Any, pd.DataFrame], allow_failed=False) -> dict[typing.Any, pd.DataFrame]: ...
def code_for_presence(markers: pd.DataFrame|dict[typing.Any, pd.DataFrame], allow_failed=False) -> pd.DataFrame|dict[typing.Any, pd.DataFrame]:
    if isinstance(markers,dict):
        for i in markers:
            markers[i] = _code_for_presence_impl(markers[i], f'{i}_', allow_failed)
    else:
        markers = _code_for_presence_impl(markers,'', allow_failed)
    return markers

def _code_for_presence_impl(markers: pd.DataFrame, lbl_extra:str, allow_failed=False) -> pd.DataFrame:
    new_col_lbl = f'marker_{lbl_extra}presence'
    markers.insert(len(markers.columns),
        new_col_lbl,
        True if allow_failed else markers[[c for c in markers.columns if c not in ['frame_idx']]].notnull().all(axis='columns')
    )
    markers = markers[['frame_idx',new_col_lbl]] if 'frame_idx' in markers else markers[[new_col_lbl]]
    markers = markers.astype({new_col_lbl: bool}) # ensure the new column is bool
    return markers

def expand_detection(markers: pd.DataFrame, fill_value):
    if 'frame_idx' in markers.columns:
        min_fr_idx = markers['frame_idx'].min()
        max_fr_idx = markers['frame_idx'].max()
        new_index = pd.Index(range(min_fr_idx,max_fr_idx+1), name='frame_idx')
        return markers.set_index('frame_idx').reindex(new_index, fill_value=fill_value).reset_index()
    else:
        if markers.index.name!='frame_idx':
            raise ValueError(f'It was expected that the name of the index is "frame_idx". It was "{markers.index.name}" instead. This may mean this dataframe does not contain the expected information. Cannot continue.')
        min_fr_idx = markers.index.min()
        max_fr_idx = markers.index.max()
        new_index = pd.Index(range(min_fr_idx,max_fr_idx+1), name='frame_idx')
        return markers.reindex(new_index, fill_value=fill_value)

def get_appearance_starts_ends(m: pd.DataFrame, max_gap_duration: int, min_duration: int):
    vals   = np.pad(m['marker_presence'].values.astype(int), (1, 1), 'constant', constant_values=(0, 0))
    d      = np.diff(vals)
    starts = np.nonzero(d == 1)[0]
    ends   = np.nonzero(d == -1)[0]
    gaps   = starts[1:]-ends[:-1]
    # fill gaps in marker detection
    gapi   = np.nonzero(gaps<=max_gap_duration)[0]
    starts = np.delete(starts,gapi+1)
    ends   = np.delete(ends,gapi)
    # remove too short
    lengths= ends-starts
    shorti = np.nonzero(lengths<min_duration)[0]
    starts = np.delete(starts,shorti)
    ends   = np.delete(ends,shorti)
    # turn first and last frames into frame_idx values
    if 'frame_idx' in m.columns:
        return m.loc[starts,'frame_idx'].to_numpy(), m.loc[ends-1,'frame_idx'].to_numpy() # NB: -1 so that ends point to last frame during which marker was last seen (and to not index out of the array)
    elif m.index.name=='frame_idx':
        return m.index[starts].to_numpy(), m.index[ends-1].to_numpy()

def get_sequence_interval(starts: dict[int,list[int]], ends: dict[int,list[int]], pattern: list[int], max_intermarker_gap_duration: int, side='start') -> np.ndarray:
    # find marker pattern (sequence of markers following in right order with gap no longer than max_intermarker_gap_duration)
    pairs: list[tuple[int,int]] = []
    for i in range(len(ends[pattern[0]])):
        end_idx = i
        for j in range(len(pattern)-1):
            if end_idx is None:
                break
            end     = ends[pattern[j]][end_idx]
            gaps    = starts[pattern[j+1]]-end
            end_idx = get_smallest_gap_end(gaps,max_intermarker_gap_duration)
        if end_idx is not None:
            pairs.append((starts[pattern[0]][i], ends[pattern[-1]][end_idx]))

    idx = 0 if side=='start' else 1
    return np.array([p[idx] for p in pairs])

def get_smallest_gap_end(gaps: np.ndarray, max_intermarker_gap_duration: int):
    gapi = np.nonzero(np.logical_and(gaps>=0, gaps<=max_intermarker_gap_duration))[0]
    if gapi.size:
        # if there are multiple that qualify, take the smallest gap
        mini = np.argmin(gaps[gapi])
        return gapi[mini]
    return None



import numpy as np
import pandas as pd
import cv2
import pathlib
import math
import typing


from . import data_files, drawing, marker, transforms

class Coordinate(typing.NamedTuple):
    x: float = 0.
    y: float = 0.

class Plane:
    default_ref_image_name = 'reference_image.png'

    def __init__(self,
                 markers                : str|pathlib.Path|pd.DataFrame,                            # if str or Path: file from which to read markers. Else direction N_markerx4 array. Should contain centers of markers
                 marker_size            : float,                                                    # in "unit" units
                 plane_size             : Coordinate,                                               # in "unit" units

                 aruco_dict_id                                          = cv2.aruco.DICT_4X4_250,
                 marker_border_bits                                     = 1,
                 marker_pos_scale_fac                                   = 1.,                       # scale factor for marker positions in the markers input argument
                 unit                   : str                           = None,                     # Unit in which measurements (marker size and positions for instance) are expressed. Purely informational
                 package_to_read_from   : str                           = None,                     # if provided, reads marker file from specified package's resources
                 ref_image_store_path   : str|pathlib.Path              = None,
                 ref_image_size         : int                           = 1920,                     # largest dimension
                 min_num_markers        : int                           = 3                         # minimum number of markers for gaze to be mapped to this plane
                 ):

        self.marker_size                                    = marker_size
        # marker positions
        self.markers            : dict[int,marker.Marker]   = {}
        self._all_marker_ids    : list[int]                 = []
        self.plane_size                                     = plane_size
        self.bbox               : list[float]               = [0., 0., self.plane_size.x, self.plane_size.y]
        self._origin            : Coordinate                = Coordinate(0., 0.)

        # marker specs
        self.aruco_dict_id                                  = aruco_dict_id
        self.aruco_dict                                     = cv2.aruco.getPredefinedDictionary(self.aruco_dict_id)
        self.marker_border_bits                             = marker_border_bits
        self.unit                                           = unit

        # processing specs
        self.min_num_markers                                = min_num_markers

        # prep markers
        self._load_markers(markers, marker_pos_scale_fac, package_to_read_from)

        # get reference image of plane
        if ref_image_store_path:
            ref_image_store_path = pathlib.Path(ref_image_store_path)

        # get image (always create reference image, to be safe (settings may have changed))
        img = self._store_reference_image(ref_image_store_path, ref_image_size)

        self._ref_image_size                                = ref_image_size
        self._ref_image_cache   : dict[int, np.ndarray]     = {ref_image_size: img}

    def set_origin(self, origin: Coordinate):
        # change from current origin
        offset = np.array(origin)-self._origin

        for i in self.markers:
            self.markers[i].shift(-offset)

        self.bbox[0] -= offset[0]
        self.bbox[2] -= offset[0]
        self.bbox[1] -= offset[1]
        self.bbox[3] -= offset[1]

        self._origin = origin

    def get_ref_image(self, im_size: int=None, as_RGB=False) -> np.ndarray:
        if im_size is None:
            im_size = self._ref_image_size
        if not isinstance(im_size,int):
            raise TypeError(f'width input should be an int, not {type(im_size)}')
        # check we have the image, if not, add to cache
        if im_size not in self._ref_image_cache:
            scale = float(im_size)/self._ref_image_size
            self._ref_image_cache[im_size] = cv2.resize(self._ref_image_cache[self._ref_image_size], None, fx=scale, fy=scale, interpolation = cv2.INTER_AREA)

        # return
        if as_RGB:
            return self._ref_image_cache[im_size][:,:,[2,1,0]].copy()
        else:
            # OpenCV's BGR
            return self._ref_image_cache[im_size].copy()

    def draw(self, img, x, y, sub_pixel_fac=1, color=None, size=6):
        if not math.isnan(x):
            xy = transforms.to_image_pos(x, y, self.bbox, [img.shape[1], img.shape[0]])
            if color is None:
                drawing.openCVCircle(img, xy, 8, (0,255,0), -1, sub_pixel_fac)
                color = (0,0,0)
            drawing.openCVCircle(img, xy, size, color, -1, sub_pixel_fac)

    def _load_markers(self, markers: str|pathlib.Path|pd.DataFrame, marker_pos_scale_fac: float, package_to_read_from: str|None):
        from . import aruco
        # read in aruco marker positions
        markerHalfSizeMm  = self.marker_size/2.

        if isinstance(markers, pd.DataFrame):
            marker_pos = markers
        else:
            marker_pos = data_files.read_coord_file(markers, package_to_read_from)
        if marker_pos is None:
            raise RuntimeError(f"No markers could be read from the file {markers}, check it exists and contains markers")

        # keep track of all IDs so we can check for duplicates
        self._all_marker_ids = marker_pos.index.to_list()
        marker_dict_size = aruco.get_dict_size(self.aruco_dict_id)
        for m_id in self._all_marker_ids:
            if m_id>=marker_dict_size:
                raise ValueError(f'This plane is set up using the dictionary {aruco.dict_id_to_str[self.aruco_dict_id]} which only has {marker_dict_size} markers, which means that valid IDs are 0-{marker_dict_size-1}. However, this plane is configured to contain a marker number {m_id} that is not a valid marker for this dictionary.')

        # turn into marker objects
        marker_pos.x *= marker_pos_scale_fac
        marker_pos.y *= marker_pos_scale_fac

        for idx, row in marker_pos.iterrows():
            c   = row[['x','y']].values
            # rotate markers (negative because plane coordinate system)
            rot = row[['rotation_angle']].values[0] if 'rotation_angle' in row else 0.
            rotr= -math.radians(rot)
            R   = np.array([[math.cos(rotr), math.sin(rotr)], [-math.sin(rotr), math.cos(rotr)]])
            # top left first, and clockwise: same order as detected ArUco marker corners
            tl = c + np.matmul(R, np.array([-markerHalfSizeMm, -markerHalfSizeMm]))
            tr = c + np.matmul(R, np.array([ markerHalfSizeMm, -markerHalfSizeMm]))
            br = c + np.matmul(R, np.array([ markerHalfSizeMm,  markerHalfSizeMm]))
            bl = c + np.matmul(R, np.array([-markerHalfSizeMm,  markerHalfSizeMm]))

            self.markers[idx] = marker.Marker(idx, c, corners=[tl, tr, br, bl], rot=rot)

    def get_marker_IDs(self) -> dict[str|int,list[marker.MarkerID]]:
        return {'plane': [marker.MarkerID(m_id, self.aruco_dict_id) for m_id in self._all_marker_ids]}

    def get_aruco_board(self) -> cv2.aruco.Board:
        from . import aruco
        board_corner_points = []
        ids = []
        for key in self.markers:
            ids.append(key)
            marker_corner_points = np.vstack(self.markers[key].corners).astype('float32')
            board_corner_points.append(marker_corner_points)
        return aruco.create_board(board_corner_points, ids, self.aruco_dict)

    def get_plane_setup(self):
        from . import aruco
        return aruco.PlaneSetup(plane = self,
                                aruco_detector_params = {
                                    'markerBorderBits': self.marker_border_bits
                                },
                                min_num_markers = self.min_num_markers)

    def _store_reference_image(self, path: pathlib.Path, im_size: int) -> np.ndarray:
        # get image with markers
        bbox_extents = [self.bbox[2]-self.bbox[0], math.fabs(self.bbox[3]-self.bbox[1])]  # math.fabs to deal with bboxes where (-,-) is bottom left
        aspect_ratio = bbox_extents[0]/bbox_extents[1]
        if aspect_ratio>1:
            width       = im_size
            height      = math.ceil(im_size/aspect_ratio)
        else:
            width       = math.ceil(im_size*aspect_ratio)
            height      = im_size

        img = np.zeros((height, width), np.uint8)
        img[:] = 255
        # collect all markers
        corner_points = []
        ids = []
        # for checking if markers fit
        x_margin = bbox_extents[0]/width /5     # ignore .2 pixel or less
        y_margin = bbox_extents[1]/height/5     # ignore .2 pixel or less
        for key in self.markers:
            ids.append(key)
            corners = np.vstack(self.markers[key].corners).astype('float32')
            # check we're on the plane
            if np.any(corners[:,0]<-x_margin) or np.any(corners[:,0]>self.plane_size.x+x_margin) or \
               np.any(corners[:,1]<-y_margin) or np.any(corners[:,1]>self.plane_size.y+y_margin):
                center  = ", ".join(map(lambda x: f"{x:.4f}",self.markers[key].center))
                corners = [", ".join(map(lambda x: f"{x:.4f}",c)) for c in corners]
                plane_corners = [", ".join(map(lambda x: f"{x:.4f}",c)) for c in (self.bbox[:2],self.bbox[2:])]
                raise ValueError(f'Marker {key} with center positioned at ({center}), size {self.marker_size:.4f} and rotation {self.markers[key].rot:.1f} deg would\nhave its corners at ({corners[0]}), ({corners[1]}), ({corners[2]}), and ({corners[3]}),\nwhich is outside the defined plane which ranges from ({plane_corners[0]}) to ({plane_corners[1]}). Ensure all\nsizes and positions are in the same unit (e.g. mm) and check the marker position csv file, marker size and plane size.')
            corner_points.append(corners)

        # get info about marker positions on the board
        corner_points = np.dstack(corner_points)
        corner_points = corner_points-np.expand_dims(np.array([self.bbox[:2]]),2).astype('float32')

        # get position and size of marker in the generated image
        corner_points[:,0,:] = corner_points[:,0,:]/bbox_extents[0] * float(img.shape[1])
        corner_points[:,1,:] = corner_points[:,1,:]/bbox_extents[1] * float(img.shape[0])

        # get marker size
        pix_sz = np.vstack((np.hypot(corner_points[0,0,:]-corner_points[1,0,:], corner_points[0,1,:]-corner_points[1,1,:]),
                            np.hypot(corner_points[1,0,:]-corner_points[2,0,:], corner_points[1,1,:]-corner_points[2,1,:]))).T
        # marker should be square
        pix_sz = np.round(np.min(pix_sz,1)).astype('int')

        # place markers
        for i,sz,pos in zip(ids,pix_sz,np.moveaxis(corner_points, -1, 0)):
            # make marker
            marker_image = np.zeros((sz, sz), dtype=np.uint8)
            marker_image = self.aruco_dict.generateImageMarker(i, sz, marker_image, self.marker_border_bits)

            # put in image
            if pos[0,1]==pos[1,1] and pos[1,0]==pos[2,0] and pos[0,0]<pos[1,0]:
                # marker is aligned to image axes and not rotated, just blit
                ori = np.round(pos[0,:]).astype('int')
                img[ori[1]:ori[1]+sz, ori[0]:ori[0]+sz] = marker_image
                continue

            # set up affine transformation for placing marker in image
            in_corners = np.array([[-.5, -.5],
                                   [marker_image.shape[1]-.5, -.5],
                                   [marker_image.shape[1]-.5, marker_image.shape[0]-.5]]).astype('float32')
            transformation = cv2.getAffineTransform(in_corners, pos[:3,:])

            # perform affine transformation (i.e. rotate marker)
            img = cv2.warpAffine(marker_image, transformation, (img.shape[1], img.shape[0]), img, flags=cv2.INTER_LINEAR, borderMode=cv2.BORDER_TRANSPARENT)

        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        if path:
            cv2.imwrite(path, img)

        return img


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pathlib

from .. import gaze_worldref, marker, naming

# assign intervals to targets based on distance
def distance(
        targets: dict[int, np.ndarray],
        fixations: str|pathlib.Path|pd.DataFrame,
        do_global_shift = True,
        max_dist_fac = .5
    ) -> tuple[pd.DataFrame, pd.DataFrame|None]:

    # read input if needed
    if not isinstance(fixations,pd.DataFrame):
        fixations = pd.read_csv(fixations,sep='\t',index_col=False)

    # for each target, find closest fixation
    minDur      = 100       # ms
    used        = np.zeros((fixations['start'].size),dtype='bool')
    selected    = np.empty((len(targets),),dtype='int')
    selected[:] = -999

    t_x = np.array([targets[t][0] for t in targets])
    t_y = np.array([targets[t][1] for t in targets])
    off_f_x = off_f_y = off_t_x = off_t_y = 0.
    if do_global_shift:
        # first, center the problem. That means determine and remove any overall shift from the
        # data and the targets, to improve robustness of assigning fixations points to targets.
        # Else, if all data is e.g. shifted up by more than half the distance between
        # validation targets, target assignment would fail
        off_f_x = fixations['xpos'].median()
        off_f_y = fixations['ypos'].median()
        off_t_x = t_x.mean()
        off_t_y = t_y.mean()

    # we furthermore do not assign a fixation to a target if the closest fixation is more than
    # max_dist_fac the intertarget distance away
    # determine intertarget distance, if possible
    dist_lim = np.inf
    if len(t_x)>1:
        # arbitrarily take first target and find closest target to it
        dist = np.hypot(t_x[0]-t_x[1:], t_y[0]-t_y[1:])
        min_dist = dist.min()
        if min_dist > 0:
            dist_lim = min_dist*max_dist_fac

    for i,t in zip(range(len(targets)),targets):
        if np.all(used):
            # all fixations used up, can't assign anything to remaining targets
            continue
        # select fixation
        dist        = np.hypot(fixations['xpos']-off_f_x-(targets[t][0]-off_t_x), fixations['ypos']-off_f_y-(targets[t][1]-off_t_y))
        dist[used]  = np.inf                    # make sure fixations already bound to a target are not used again
        dist[fixations['dur']<minDur] = np.inf  # make sure that fixations that are too short are not selected
        iFix        = np.argmin(dist)
        if dist[iFix]<=dist_lim:
            selected[i] = iFix
            used[iFix]  = True

    # prep return values
    selected_intervals = pd.DataFrame(columns=['xpos','ypos','startT','endT'])  # sets which columns to copy
    selected_intervals.index.name = 'target'
    for i,t in zip(range(len(targets)),targets):
        if selected[i]==-999:
            continue
        selected_intervals.loc[t] = fixations.iloc[selected[i]]
    other_intervals = fixations.loc[np.logical_not(used),['xpos','ypos','startT']]
    return selected_intervals, None if other_intervals.empty else other_intervals

def dynamic_markers(
        marker_observations_per_target: dict[int, pd.DataFrame],    # {target: dataframe}
        markers_per_target: dict[int,list[marker.MarkerID]],        # {target: [marker,...]}, one or multiple markers per target
        timestamps_file: str|pathlib.Path,
        episode: list[int],
        skip_first_duration: float,
        max_gap_duration: int,
        min_duration: int
    ) -> tuple[pd.DataFrame, None]:
    # also need frame timestamps because the intervals returned by this function should be expressed in recording time, not as frame indices.
    timestamps  = pd.read_csv(timestamps_file, delimiter='\t', index_col='frame_idx')
    ts_col      = 'timestamp_stretched' if 'timestamp_stretched' in timestamps else 'timestamp'

    # make local copy of marker_observations, containing only the current episode
    marker_observations_per_target = {t:mo.loc[episode[0]:episode[1],:] for t,mo in marker_observations_per_target.items()}
    # check we have data for at least one of the markers for a given target
    for t in marker_observations_per_target:
        if marker_observations_per_target[t].empty:
            missing_str  = '\n- '.join([marker.marker_ID_to_str(m) for m in markers_per_target[t]])
            raise RuntimeError(f'None of the markers for target {t} were observed during the episode from frame {episode[0]} to frame {episode[1]}:\n- {missing_str}')

    # marker presence signal only contains marker detections (True). We need to fill the gaps in between detections with False (not detected) so we have a continuous signal without gaps
    marker_observations_per_target = {t: marker.expand_detection(marker_observations_per_target[t], fill_value=False) for t in marker_observations_per_target}

    # for each target, see when it is presented using the marker presence signal
    selected_intervals = pd.DataFrame(columns=['startT','endT'])
    selected_intervals.index.name = 'target'
    for t in marker_observations_per_target:
        start, end = marker.get_appearance_starts_ends(marker_observations_per_target[t], max_gap_duration, min_duration)
        if start.size==0:
            continue
        # in case there are multiple (e.g. spotty detection), choose longest
        durs = np.array(end)-np.array(start)+1
        maxi = np.argmax(durs)
        ts = timestamps.loc[[start[maxi], end[maxi]],ts_col].to_numpy()
        ts[0] += skip_first_duration
        if ts[0]>=ts[1]:
            continue
        selected_intervals.loc[t] = ts
    return selected_intervals, None

def plot(
        selected_intervals: pd.DataFrame,
        other_intervals: pd.DataFrame|None,
        targets: dict[int, np.ndarray],
        gazes: str|pathlib.Path|dict[int, list[gaze_worldref.Gaze]],
        episode: list[int],
        output_directory: str|pathlib.Path,
        filename_stem: str = naming.fixation_assignment_prefix,
        iteration = 0,
        background_image: tuple[np.ndarray, list[float]] = None,    # (image, extent in mm [l r t b])
        plot_limits: list[list[float]] = None
    ):
    output_directory = pathlib.Path(output_directory)
    # if we do not have x and y positions for the gaze intervals, make them
    if 'xpos' not in selected_intervals.columns or (other_intervals is not None and 'xpos' not in other_intervals.columns):
        # read input if needed
        if not isinstance(gazes,dict):
            gazes = gaze_worldref.read_dict_from_file(gazes)
        if 'xpos' not in selected_intervals.columns:
            samples_per_frame = {k:v for (k,v) in gazes.items() if k>=episode[0] and k<=episode[1]}
            has_ray = np.any(np.logical_not(np.isnan([s.gazePosPlane2D_vidPos_ray for v in gazes.values() for s in v])))
            field = 'gazePosPlane2D_vidPos_ray' if has_ray else 'gazePosPlane2D_vidPos_homography'
            for t in selected_intervals.index:
                st,et = selected_intervals.loc[t,['startT','endT']].to_numpy()
                data = [getattr(s,field) for v in samples_per_frame.values() for s in v if s.timestamp>=st and s.timestamp<=et]
                if data:
                    gaze = np.vstack(data)
                    selected_intervals.loc[t,['xpos','ypos']] = np.nanmedian(gaze,axis=0)
        if other_intervals is not None and 'xpos' not in other_intervals.columns:
            for t in other_intervals.index:
                st,et = other_intervals.loc[t,['startT','endT']].to_numpy()
                data = [getattr(s,field) for v in samples_per_frame.values() for s in v if s.timestamp>=st and s.timestamp<=et]
                if data:
                    gaze = np.vstack(data)
                    other_intervals.loc[t,['xpos','ypos']] = np.nanmedian(gaze,axis=0)
    # make plot of data overlaid on poster, and show for each target which interval was selected
    # prep data
    if other_intervals is not None:
        all_intervals = pd.concat((selected_intervals.set_index('startT'), other_intervals.set_index('startT')), join='inner').sort_index()
    else:
        all_intervals = selected_intervals.set_index('startT').sort_index()

    f       = plt.figure(dpi=300)
    imgplot = plt.imshow(background_image[0],extent=background_image[1],alpha=.5)
    # draw all intervals
    plt.plot(all_intervals['xpos'],all_intervals['ypos'],'b-')
    plt.plot(all_intervals['xpos'],all_intervals['ypos'],'go')
    # draw target matching
    for t, row in selected_intervals.iterrows():
        plt.plot([row['xpos'], targets[t][0]], [row['ypos'], targets[t][1]],'r-')

    # cosmetics
    plt.xlabel('mm')
    plt.ylabel('mm')
    if plot_limits is not None:
        plt.xlim(plot_limits[0])
        plt.ylim(plot_limits[1])
    plt.gca().invert_yaxis()

    f.savefig(output_directory / f'{filename_stem}_interval_{iteration+1:02d}.png')
    plt.close(f)

def to_tsv(
        selected_intervals: pd.DataFrame,
        output_directory: str|pathlib.Path,
        filename_stem: str = naming.fixation_assignment_prefix,
        iteration = 0,
    ):
    output_directory = pathlib.Path(output_directory)
    # store selected intervals
    selected_intervals = selected_intervals \
                            .drop(columns=[c for c in ('xpos','ypos') if c in selected_intervals.columns]) \
                            .rename(columns={'startT': 'start_timestamp', 'endT': 'end_timestamp'})
    selected_intervals.insert(0, 'marker_interval', iteration+1)

    selected_intervals.to_csv(
        output_directory / f'{filename_stem}.tsv',
        mode='w' if iteration==0 else 'a',
        header=iteration==0,
        sep='\t', na_rep='nan', float_format="%.3f")


import pathlib
import importlib.resources
import shutil
import json
import math
import pandas as pd
from collections import defaultdict

from ..config import get_markers, get_targets
from ... import aruco, marker

def deploy_setup_and_script(output_dir: str|pathlib.Path, overwrite=False) -> list[str]:
    output_dir = pathlib.Path(output_dir)
    if not output_dir.is_dir():
        raise RuntimeError(f'The requested directory "{output_dir}" does not exist')

    # copy over all files
    not_copied: list[str] = []
    for r in ['markerPositions.csv', 'targetPositions.csv', 'setup.json', 'stim.py']:
        out_file = output_dir/r
        if out_file.exists() and not overwrite:
            not_copied.append(r)
            continue
        with importlib.resources.path(__package__, r) as p:
            shutil.copyfile(p, out_file)
    return not_copied

def _get_extent(extent: float, distance: float, psychopy_unit: str) -> float:
    match psychopy_unit:
        case 'cm':
            return extent
        case 'deg' | 'degFlatPos':
            # sizes are not corrected for flat screen when unit is degFlatPos, so same as deg
            # for factor, see comment in _get_position
            return extent * distance*0.017455
        case _:
            raise ValueError(f'PsychoPy unit {psychopy_unit} is not understood')

def _get_position(position: tuple[float,float], distance: float, psychopy_unit: str) -> tuple[float,float]:
    match psychopy_unit:
        case 'cm':
            return position
        case 'deg':
            # 1 deg (centered) at 1 cm is approximately pi/180 (2*tand(.5)), seems to be the logic for PsychoPy
            # this is then apparently hardcoded as 0.017455, which is rounded off a bit wrong...
            fac = distance * 0.017455
            return tuple((p*fac for p in position))
        case 'degFlatPos':
            # positions corrected for flat screen, sizes not
            x,y = (math.radians(x) for x in position)
            return (math.hypot(distance, math.tan(y) * distance) * math.tan(x),
                    math.hypot(distance, math.tan(x) * distance) * math.tan(y))
        case _:
            raise ValueError(f'PsychoPy unit {psychopy_unit} is not understood')

def setup_to_automatic_coding(config_dir: str|pathlib.Path|None=None, file_name: str='setup.json') -> dict[str, int|list[int]]:
    # load config used to run validation
    # if no config dir specified, load default
    if config_dir is not None:
        with open(pathlib.Path(config_dir)/file_name, 'r') as f:
            setup = json.load(f)
    else:
        with importlib.resources.open_text(__package__, file_name) as f:
            setup = json.load(f)

    out = setup["validation"]["segment_marker"]
    out["border_bits"] = setup["segment_marker"]["border_bits"]
    return out

def setup_to_plane_config(output_dir: str|pathlib.Path, config_dir: str|pathlib.Path|None=None, file_name: str='setup.json') -> dict[str,list[marker.MarkerID]]:
    output_dir = pathlib.Path(output_dir)
    if not output_dir.is_dir():
        raise RuntimeError(f'The requested directory "{output_dir}" does not exist')

    # load config used to run validation
    # if no config dir specified, load default
    if config_dir is not None:
        with open(pathlib.Path(config_dir)/file_name, 'r') as f:
            setup = json.load(f)
    else:
        with importlib.resources.open_text(__package__, file_name) as f:
            setup = json.load(f)

    # get units in which positions are expressed, and viewing distance which may be needed for converting them
    marker_units = setup["validation"]["markers"]["units"] if "units" in setup["validation"]["markers"] else setup["screen"]["units"]
    target_units = setup["validation"]["targets"]["units"] if "units" in setup["validation"]["targets"] else setup["screen"]["units"]
    dist = setup["screen"]["viewing_distance"]
    # load target and marker positions (default ones from package if config dir is not specified)
    target_positions = get_targets(config_dir, setup["validation"]["targets"]["file"], __package__)
    marker_positions = get_markers(config_dir, setup["validation"]["markers"]["file"], __package__)
    # convert positions to cm
    target_positions[['x','y']] = target_positions.apply(lambda r: _get_position((r['x'],r['y']), dist, target_units), axis=1, result_type='expand')
    marker_positions[['x','y']] = marker_positions.apply(lambda r: _get_position((r['x'],r['y']), dist, marker_units), axis=1, result_type='expand')
    # negate y as positive y is upward for PsychoPy script but downward for our logic
    target_positions['y'] = -target_positions['y']
    marker_positions['y'] = -marker_positions['y']
    # Marker and target positions are centers, and relative to center of screen.
    # Our positions should have (0,0) in the top-left. Shift (taking sizes into account so who markers and targets fit on the reference image)
    t_size= _get_extent(setup["validation"]["targets"]["look"]["diameter_max"], dist, target_units)
    m_size= _get_extent(setup["validation"]["markers"]["size"]                , dist, marker_units)
    min_t = target_positions[['x','y']].min()-t_size/2
    min_m = marker_positions[['x','y']].min()-m_size/2
    x_min = min(min_t['x'],min_m['x'])
    y_min = min(min_t['y'],min_m['y'])
    target_positions['x']-=x_min
    marker_positions['x']-=x_min
    target_positions['y']-=y_min
    marker_positions['y']-=y_min

    # get plane size
    max_t = target_positions[['x','y']].max()+t_size/2
    max_m = marker_positions[['x','y']].max()+m_size/2
    x_max = max(max_t['x'],max_m['x'])
    y_max = max(max_t['y'],max_m['y'])

    # write appropriate validationSetup.txt file
    valSetup = {
        'distance': dist,
        'mode': 'cm',
        'arucoDictionary': setup["aruco"]["dict"],
        'markerBorderBits': setup["aruco"]["border_bits"],
        'markerSide': m_size,
        'markerPosFile': 'markerPositions_converted.csv',
        'targetPosFile': 'targetPositions_converted.csv',
        'targetType': 'Thaler',
        'targetDiameter': t_size,
        'showGrid': 0,
        'gridCols': x_max,
        'gridRows': y_max,
        'minNumMarkers': 3,
        'centerTarget': setup["validation"]["targets"]["center_target"],
        'referencePosterSize': 1920
    }

    # add marker indicator(s) as columns to target file
    for i,_ in enumerate(setup["validation"]["markers"]["replace_IDs"]):
        IDs = target_positions.index.to_numpy()+setup["validation"]["markers"]["replace_ID_start"]+i*setup["validation"]["markers"]["replace_ID_offset"]
        target_positions[f'marker_{i}'] = IDs

    # store everything to files
    with open(output_dir/'validationSetup.txt','w') as f:
        for key,value in valSetup.items():
            f.write(f'{key} = {value}\n')
    target_positions.to_csv(output_dir/valSetup['targetPosFile'], float_format='%.8f')
    marker_positions.to_csv(output_dir/valSetup['markerPosFile'], float_format='%.8f')

    # last, get segmentation markers used for start and end of validation interval
    segmentation_markers: dict[str,list[marker.MarkerID]] = {}
    aruco_dict_id = aruco.str_to_dict_id(setup["aruco"]["dict"])
    for s,o in zip(('start_IDs','end_IDs'),('start_markers','end_markers')):
        segmentation_markers[o] = [marker.MarkerID(m, aruco_dict_id) for m in setup["validation"]["segment_marker"][s]]
    segmentation_markers['marker_border_bits'] = setup["aruco"]["border_bits"]
    return segmentation_markers


# for analysis
def get_marker_observations(validation_plane: 'validation.Plane', working_dir: pathlib.Path) -> tuple[dict[int, pd.DataFrame], dict[int,list[marker.MarkerID]]]:
    # organize markers
    markers_per_target: dict[int,list[marker.MarkerID]] = defaultdict(list)
    for m in validation_plane.dynamic_markers:
        t = validation_plane.dynamic_markers[m][0]
        markers_per_target[t].append(marker.MarkerID(m, validation_plane.aruco_dict_id))
    markers_per_target = dict(markers_per_target)   # get rid of defaultdict now its no longer needed so we get normal indexing

    # determine what marker files to read
    all_marker_ids = [m for ms in markers_per_target for m in markers_per_target[ms]]
    # for each target, check at least one of the marker files exists
    for t in markers_per_target:
        missing = [not marker.get_file_name(m.m_id, m.aruco_dict_id, working_dir).is_file() for m in markers_per_target[t]]
        if all(missing):
            file_missing = [marker.get_file_name(m.m_id, m.aruco_dict_id, None) for m in markers_per_target[t]]
            missing_str  = '\n- '.join(file_missing)
            raise FileNotFoundError(f'None of the marker files for target {t} were found:\n- {missing_str}')
        # remove missing from list of markers to load
        if any(missing):
            for i,m in enumerate(missing):
                if not m:
                    continue
                all_marker_ids.remove(markers_per_target[t][i])

    # load all markers and recode so we just have a boolean indicating when markers are present
    marker_observations = {m: marker.read_dataframe_from_file(m.m_id, m.aruco_dict_id, working_dir).set_index('frame_idx') for m in all_marker_ids}
    marker_observations = {m: marker.code_for_presence(marker_observations[m], allow_failed=True) for m in marker_observations if not marker_observations[m].empty}

    # target presentations may be encoded by multiple markers simultaneously
    # merge all markers for the target to be more robust to choppy detection
    marker_observations_per_target: dict[int, pd.DataFrame] = {}
    for t in markers_per_target:
        for m in markers_per_target[t]:
            if m not in marker_observations or marker_observations[m].empty:
                continue
            if t not in marker_observations_per_target:
                marker_observations_per_target[t] = marker_observations[m]
            else:
                marker_observations_per_target[t] = marker_observations_per_target[t].combine_first(marker_observations[m])

    return marker_observations_per_target, markers_per_target